using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {

	[SupportedExtension("bsp")]
	public partial class Level : INamedResource, IBoundingBoxed {

		#region Member Variables

		/// <summary>The stream we load the data from.</summary>
		private readonly Stream BaseStream;

		/// <summary>List of raw directory entries.</summary>
		private Entry[] Entries;

		/// <summary>The raw, compressed, visibility list data.</summary>
		private byte[] CompressedVisibilityLists;

		#endregion

		#region Properties


		private readonly Vector3[] vertices;
		/// <summary>Gets the vertices for the level.</summary>
		[Category("Geometry")]
		[Description("All of the vertices in the level.")]
		public Vector3[] Vertices {
			get { return this.vertices; }
		}

		private readonly int[] edgeList;
		/// <summary>Gets the list of edges for the level.</summary>
		[Category("Geometry")]
		[Description("List of edges as referenced by the individual faces.")]
		public int[] EdgeList {
			get { return this.edgeList; }
		}


		private readonly ushort[] faceList;
		/// <summary>Gets the level's face list.</summary>
		[Category("Geometry")]
		[Description("List of faces as referenced by the leaves.")]
		public ushort[] FaceList {
			get { return this.faceList; }
		}

		/// <summary>Gets true if the level contains visibility lists.</summary>
		[Description("True if the level contains visibility lists.")]
		public bool HasVisibilityLists {
			get { return this.CompressedVisibilityLists.Length > 0; }
		}

		/// <summary>Gets the bounding box for the entire level.</summary>
		public BoundingBox Bounds {
			get { return this.nodes[0].Bounds; }
		}

		private string name;
		[Browsable(false)]
		public string Name { get { return this.name; } set { this.name = value; } }

		[Browsable(false)]
		public bool NamedInternally { get { return false; } }

		#endregion

		private readonly ResourceLoader Resources;


		public Level(ResourceLoader resourceLoader, Stream data) {
			this.BaseStream = data;
			this.Resources = resourceLoader;

			BinaryReader LevelReader = new BinaryReader(data);

			if (Encoding.ASCII.GetString(LevelReader.ReadBytes(4)) != "IBSP") {
				LevelReader.BaseStream.Seek(-4, SeekOrigin.Current);
			}

			this.version = new  Version((Version.Game)LevelReader.ReadUInt32());

			// Load all of the entry data:
			this.Entries = new Entry[this.FileVersion.EntryCount];
			for (int i = 0; i < this.Entries.Length; ++i) {
				this.Entries[i] = new Entry(this);
			}

			// Load the individial parts:
			Entry WorkingEntry;
			BinaryReader WorkingReader;

			#region Vertices
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.Vertices)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.Vertices)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.vertices = new Vector3[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.Vertices)];
				for (int i = 0; i < this.vertices.Length; ++i) {
					this.vertices[i] = new Vector3(WorkingReader.BaseStream);
				}
			}
			#endregion

			#region Edges
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.Edges)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.Edges)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.edges = new Edge[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.Edges)];
				for (int i = 0; i < this.edges.Length; ++i) {
					this.edges[i] = new Edge(WorkingReader.ReadUInt16(), WorkingReader.ReadUInt16());
				}
			}
			#endregion

			#region Edge List
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.EdgeList)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.EdgeList)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.edgeList = new int[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.EdgeList)];
				for (int i = 0; i < this.edgeList.Length; ++i) {
					this.edgeList[i] = WorkingReader.ReadInt32();
				}
			}
			#endregion

			#region Models
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.Models)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.Models)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.models = new Model[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.Models)];
				for (int i = 0; i < this.models.Length; ++i) this.models[i] = new Model(this, WorkingReader.BaseStream, i);
			}
			#endregion

			#region Mip Texture List
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.MipTextures)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.MipTextures)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.textures = new MipTexture[WorkingReader.ReadInt32()];
				long[] TextureOffsets = new long[this.textures.Length];
				for (int i = 0; i < this.textures.Length; ++i) {
					TextureOffsets[i] = WorkingReader.ReadInt32();
				}

				for (int i = 0; i < this.textures.Length; ++i) {
					WorkingReader.BaseStream.Seek(TextureOffsets[i] + WorkingEntry.Offset, SeekOrigin.Begin);
					this.textures[i] = new MipTexture(this, WorkingReader.BaseStream);
				}
			}
			#endregion

			#region Texture Information
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.TextureInfo)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.TextureInfo)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.textureInformation = new TextureInfo[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.TextureInfo)];
				for (int i = 0; i < this.textureInformation.Length; ++i) {
					this.textureInformation[i] = new TextureInfo(this, WorkingReader.BaseStream);
				}
			}
			#endregion

			#region Planes
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.Planes)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.Planes)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.planes = new Plane[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.Planes)];
				for (int i = 0; i < this.planes.Length; ++i) {
					this.planes[i] = new Plane(WorkingReader.BaseStream);

				}
			}
			#endregion

			#region Nodes
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.Nodes)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.Nodes)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.nodes = new Node[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.Nodes)];
				for (int i = 0; i < this.nodes.Length; ++i) {
					this.Nodes[i] = new Node(this, WorkingReader.BaseStream);
				}
			}
			#endregion

			#region Visibility Lists
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.VisibilityLists)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.VisibilityLists)];
				WorkingReader = WorkingEntry.GetBinaryReader();

				long InitialPosition = WorkingReader.BaseStream.Position;

				this.CompressedVisibilityLists = WorkingReader.ReadBytes((int)WorkingEntry.Length);

				if (this.version.HasVisibilityClusters) {
					WorkingReader.BaseStream.Seek(InitialPosition, SeekOrigin.Begin);
					this.visibilityClusters = new VisibilityCluster[WorkingReader.ReadUInt32()];
					for (int i = 0; i < this.visibilityClusters.Length; ++i) this.visibilityClusters[i] = new VisibilityCluster(this, WorkingReader.BaseStream);

				} else {
					this.visibilityClusters = new VisibilityCluster[0];
				}

			}
			#endregion

			#region Leaves
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.Leaves)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.Leaves)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.leaves = new Leaf[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.Leaves)];
				for (int i = 0; i < this.leaves.Length; ++i) {
					this.leaves[i] = new Leaf(this, WorkingReader.BaseStream);
					if (this.FileVersion.HasVisibilityClusters && this.leaves[i].HasVisibilityList) this.visibilityClusters[this.leaves[i].VisibilityCluster].visibleLeaves.Add(i);
				}
			}
			#endregion

			#region Face List
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.FaceList)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.FaceList)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.faceList = new ushort[WorkingEntry.Length / 2];
				for (int i = 0; i < this.faceList.Length; ++i) {
					this.faceList[i] = WorkingReader.ReadUInt16();
				}
			}
			#endregion

			#region Faces
			if (this.FileVersion.SupportsEntry(Entry.EntryTypes.Faces)) {
				WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.Faces)];
				WorkingReader = WorkingEntry.GetBinaryReader();
				this.faces = new Face[WorkingEntry.Length / this.FileVersion.GetEntrySize(Entry.EntryTypes.Faces)];
				for (int i = 0; i < this.faces.Length; ++i) {
					this.faces[i] = new Face(this, WorkingReader.BaseStream);
				}
			}
			#endregion

			#region Entities

			WorkingEntry = this.Entries[this.FileVersion.GetEntryIndex(Entry.EntryTypes.Entities)];
			WorkingReader = WorkingEntry.GetBinaryReader();
			this.entities = new EntityList(Encoding.ASCII.GetString(WorkingReader.ReadBytes((int)WorkingEntry.Length)));

			#endregion


		}

	}
}
