using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class Level {

		#region Properties

		private readonly MipTexture[] textures = new MipTexture[0];
		/// <summary>Gets the level's textures.</summary>
		[Category("Textures")]
		public MipTexture[] MipTextures {
			get { return this.textures; }
		}

		#endregion

		/// <summary>Describes a mip texture.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class MipTexture : IPicture, IMipMappedPicture {

			#region Public Properties

			private readonly string name;
			/// <summary>Gets the name of the texture.</summary>
			public string Name {
				get { return this.name; }
			}

			private readonly int width;
			/// <summary>Gets the width of the texture in pixels.</summary>
			public int Width {
				get { return this.width; }
			}

			private readonly int height;
			/// <summary>Gets the height of the texture in pixels.</summary>
			public int Height {
				get { return this.height; }
			}

			#endregion

			#region Private Members

			private readonly byte[][] Data;

			/// <summary>Base offset of the texture data in the stream.</summary>
			private readonly long BaseOffset;

			private readonly Level Level;

			#endregion


			public MipTexture(Level level, Stream stream) {
				this.Level = level;
				this.BaseOffset = stream.Position;
				BinaryReader WorkingReader = new BinaryReader(stream);
				this.name = Encoding.ASCII.GetString(WorkingReader.ReadBytes(16)).Split('\0')[0];
				this.width = WorkingReader.ReadInt32();
				this.height = WorkingReader.ReadInt32();
				int[] Offsets = new int[] { WorkingReader.ReadInt32(), WorkingReader.ReadInt32(), WorkingReader.ReadInt32(), WorkingReader.ReadInt32() };
				if (string.IsNullOrEmpty(this.name)) {
					this.width = 1;
					this.height = 1;
				}

				long OldPosition = stream.Position;
				this.Data = new byte[4][];
				for (int i = 0; i < 4; ++i) {
					int Width = this.Width >> i;
					int Height = this.Height >> i;
					stream.Position = Offsets[i] + this.BaseOffset;
					this.Data[i] = WorkingReader.ReadBytes(Width * Height);
				}
				stream.Position = OldPosition;

			}

			public override string ToString() {
				return this.Name;
			}

			#region IPicture Members

			private Palette SourcePalette {
				get {
					Palette P;
					if (!this.Level.Resources.TryGetCachedObject("gfx/palette.lmp", out P)) {
						P = new Palette(this.Level.Resources, this.Level.Resources.OpenFile("gfx/palette.lmp"));
						this.Level.Resources.CacheObject("gfx/palette.lmp", P);
					}
					return P;
				}
			}

			private ColourMap SourceColourMap {
				get {
					ColourMap C;
					if (!this.Level.Resources.TryGetCachedObject("gfx/colormap.lmp", out C)) {
						C = new ColourMap(this.Level.Resources, this.Level.Resources.OpenFile("gfx/colormap.lmp"));
						this.Level.Resources.CacheObject("gfx/colormap.lmp", C);
					}
					return C;
				}
			}


			public Picture Picture {
				get {
					return GetPicture(0);
				}
			}
		
			public Picture GetPicture(int mipLevel) {
				
				int Width = this.Width >> mipLevel;
				int Height = this.Height >> mipLevel;

				Palette P = this.SourcePalette;

				int[] Data = Array.ConvertAll<byte, int>(this.Data[mipLevel], delegate(byte b) { return (P.Colours[b] & 0xFFFFFF) | (b << 24); });

				return new Picture(Width, Height, Data);
			}

			#endregion

			
		}
	}
}