using System;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class Level {


		#region Properties

		private Version version;
		public Version FileVersion {
			get { return this.version; }
		}

		#endregion

		/// <summary>Describes the capabilities of a particular BSP file version.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Version {

			public enum Game {
				Quake = 29,
				Quake2 = 38,
			}

			public override string ToString() {
				return this.GameVersion.ToString();
			}


			private Game game;
			/// <summary>Gets the game that this particular version of the format relates to.</summary>
			public Game GameVersion {
				get { return this.game; }
			}

			public Version(Game game) {
				this.game = game;
				if (!Enum.IsDefined(typeof(Game), game)) throw new ArgumentException("Unsupported version.");
			}

			#region Entry Data

			/// <summary>Gets the number of data entries in the BSP file.</summary>
			public int EntryCount {
				get {
					switch (this.GameVersion) {
						case Game.Quake:
							return 15;
						case Game.Quake2:
							return 19;
						default:
							throw new InvalidOperationException();
					}
				}
			}

			/// <summary>Gets the index of an entry from its entry type.</summary>
			/// <remarks>Quake versions share similar entries, but the order of entries has changed hence the need to use this method.</remarks>
			public int GetEntryIndex(Entry.EntryTypes entry) {

				switch (this.GameVersion) {
					case Game.Quake:
						switch (entry) {
							case Entry.EntryTypes.Entities: return 0;
							case Entry.EntryTypes.Planes: return 1;
							case Entry.EntryTypes.MipTextures: return 2;
							case Entry.EntryTypes.Vertices: return 3;
							case Entry.EntryTypes.VisibilityLists: return 4;
							case Entry.EntryTypes.Nodes: return 5;
							case Entry.EntryTypes.TextureInfo: return 6;
							case Entry.EntryTypes.Faces: return 7;
							case Entry.EntryTypes.LightMaps: return 8;
							case Entry.EntryTypes.ClipNodes: return 9;
							case Entry.EntryTypes.Leaves: return 10;
							case Entry.EntryTypes.FaceList: return 11;
							case Entry.EntryTypes.Edges: return 12;
							case Entry.EntryTypes.EdgeList: return 13;
							case Entry.EntryTypes.Models: return 14;
							default: throw new InvalidOperationException("Entry not supported by this BSP version.");
						}
					case Game.Quake2:
						switch (entry) {
							case Entry.EntryTypes.Entities: return 0;
							case Entry.EntryTypes.Planes: return 1;
							case Entry.EntryTypes.Vertices: return 2;
							case Entry.EntryTypes.VisibilityLists: return 3;
							case Entry.EntryTypes.Nodes: return 4;
							case Entry.EntryTypes.TextureInfo: return 5;
							case Entry.EntryTypes.Faces: return 6;
							case Entry.EntryTypes.LightMaps: return 7;
							case Entry.EntryTypes.Leaves: return 8;
							case Entry.EntryTypes.FaceList: return 9;
							case Entry.EntryTypes.LeafBrushTable: return 10;
							case Entry.EntryTypes.Edges: return 11;
							case Entry.EntryTypes.EdgeList: return 12;
							case Entry.EntryTypes.Models: return 13;
							case Entry.EntryTypes.Brushes: return 14;
							case Entry.EntryTypes.BrushSides: return 15;
							case Entry.EntryTypes.Pop: return 16;
							case Entry.EntryTypes.Areas: return 17;
							case Entry.EntryTypes.AreaPortals: return 18;
							default: throw new InvalidOperationException("Entry not supported by this BSP version.");
						}
					default: throw new InvalidOperationException(); // Impossible.

				}
			}

			/// <summary>Gets the size of a single entry in bytes.</summary>
			public int GetEntrySize(Entry.EntryTypes entry) {
				switch (entry) {
					case Entry.EntryTypes.Vertices: return 12;
					case Entry.EntryTypes.Edges: return 4;
					case Entry.EntryTypes.EdgeList: return 4;
					case Entry.EntryTypes.Faces: return 20;
					case Entry.EntryTypes.Models:
						switch (this.GameVersion) {
							case Game.Quake: return 64;
							case Game.Quake2: return 48;
							default: throw new InvalidOperationException();
						}
					case Entry.EntryTypes.TextureInfo:
						switch (this.GameVersion) {
							case Game.Quake: return 40;
							case Game.Quake2: return 76;
							default: throw new InvalidOperationException();
						}
					case Entry.EntryTypes.Planes: return 20;
					case Entry.EntryTypes.Nodes:
						switch (this.GameVersion) {
							case Game.Quake: return 24;
							case Game.Quake2: return 28;
							default: throw new InvalidOperationException();
						}
					case Entry.EntryTypes.Leaves: return 28;
					case Entry.EntryTypes.FaceList: return 2;
					default: throw new InvalidOperationException();
				}
			}

			public bool SupportsEntry(Entry.EntryTypes entry) {
				switch (this.GameVersion) {
					case Game.Quake:
						switch (entry) {
							case Entry.EntryTypes.Entities:
							case Entry.EntryTypes.Planes:
							case Entry.EntryTypes.MipTextures:
							case Entry.EntryTypes.Vertices:
							case Entry.EntryTypes.VisibilityLists:
							case Entry.EntryTypes.Nodes:
							case Entry.EntryTypes.TextureInfo:
							case Entry.EntryTypes.Faces:
							case Entry.EntryTypes.LightMaps:
							case Entry.EntryTypes.ClipNodes:
							case Entry.EntryTypes.Leaves:
							case Entry.EntryTypes.FaceList:
							case Entry.EntryTypes.Edges:
							case Entry.EntryTypes.EdgeList:
							case Entry.EntryTypes.Models:
								return true;
							default:
								return false;
						}
					case Game.Quake2:
						switch (entry) {
							case Entry.EntryTypes.Entities:
							case Entry.EntryTypes.Planes:
							case Entry.EntryTypes.Vertices:
							case Entry.EntryTypes.VisibilityLists:
							case Entry.EntryTypes.Nodes:
							case Entry.EntryTypes.TextureInfo:
							case Entry.EntryTypes.Faces:
							case Entry.EntryTypes.LightMaps:
							case Entry.EntryTypes.Leaves:
							case Entry.EntryTypes.LeafBrushTable:
							case Entry.EntryTypes.Edges:
							case Entry.EntryTypes.FaceList:
							case Entry.EntryTypes.EdgeList:
							case Entry.EntryTypes.Models:
							case Entry.EntryTypes.Brushes:
							case Entry.EntryTypes.BrushSides:
							case Entry.EntryTypes.Pop:
							case Entry.EntryTypes.Areas:
							case Entry.EntryTypes.AreaPortals:
								return true;
							default:
								return false;
						}
					default: throw new InvalidOperationException();
				}
			}

			/// <summary>Gets true if the visibility list data is divided into clusters.</summary>
			public bool HasVisibilityClusters {
				get {
					switch (this.GameVersion) {
						case Game.Quake:
							return false;
						case Game.Quake2:
							return true;
						default: throw new InvalidOperationException();
					}
				}
			}

			#endregion

			#region Public Methods

			/// <summary>Gets the number of bytes per pixel in the face lightmap textures.</summary>
			public int LightMapBytesPerPixel {
				get {
					switch (this.GameVersion) {
						case Game.Quake: return 1;
						case Game.Quake2: return 3;
						default: throw new InvalidOperationException();
					}
				}
			}

			#endregion

			#region Internal Methods

			internal void BreakOnInvalidGameVersion(Game game) {
				if (game != this.GameVersion) throw new InvalidOperationException(string.Format("Unsupported game version (expected {0}).", game));
			}


			#endregion


		}
	}
}
