using System;
using System.IO;
using System.ComponentModel;
using System.Text;

namespace id.Quake.Files {

	[SupportedExtension("mdl")]
	public partial class Model : INamedResource, IBoundingBoxed {

		#region Properties

		private uint version;
		/// <summary>Gets the model file version number.</summary>
		public uint Version { get { return this.version; } }

		private Vector3 scale;
		/// <summary>Gets the model scale factors.</summary>
		[Category("Dimensions")]
		public Vector3 Scale { get { return this.scale; } }

		private Vector3 origin;
		/// <summary>Gets the model origin.</summary>
		[Category("Geometry")]
		public Vector3 Origin { get { return this.origin; } }

		private float radius;
		/// <summary>Gets the model's bounding radius.</summary>
		[Category("Dimensions")]
		public float Radius { get { return this.radius; } }

		private Vector3 offsets;
		/// <summary>Gets the eye position.</summary>
		[Category("Dimensions")]
		public Vector3 Offsets { get { return this.offsets; } }

		private int skinWidth;
		/// <summary>Gets the width of the skins in pixels.</summary>
		[Category("Skin")]
		[DisplayName("Width")]
		public int SkinWidth { get { return this.skinWidth; } }
		
		private int skinHeight;
		/// <summary>Gets the width of the skins in pixels.</summary>
		[Category("Skin")]
		[DisplayName("Height")]
		public int SkinHeight { get { return this.skinHeight; } }

		private readonly TextureCoordinate[] textureCoordinates;
		/// <summary>Gets the model's texture coordinates.</summary>
		[Category("Skin")]
		public TextureCoordinate[] TextureCoordinates {
			get { return this.textureCoordinates; }
		}

		private readonly Triangle[] triangles;
		/// <summary>Gets the model's triangles.</summary>
		[Category("Geometry")]
		public Triangle[] Triangles {
			get { return this.triangles; }
		}

		private readonly Skin[] skins;
		/// <summary>Gets the skins associated with a particular model.</summary>
		[Category("Skin")]
		public Skin[] Skins {
			get { return this.skins; }
		}

		private readonly AnimationFrame[] frames;
		/// <summary>Gets the animated frames of the model.</summary>
		[Category("Animation")]
		public AnimationFrame[] Frames {
			get { return this.frames; }
		}

		/// <summary>Gets the total number of frames.</summary>
		[Category("Animation")]
		public int TotalFrameCount {
			get {
				int Result = 0;
				foreach (AnimationFrame F in this.Frames) Result += F.Frames.Length;
				return Result;
			}
		}


		/// <summary>Gets the bounding box for the model.</summary>
		public BoundingBox Bounds {
			get { return new BoundingBox(new Vector3(-8f), new Vector3(+8f)); }
		}

		private string name;
		[Browsable(false)]
		public string Name { get { return this.name; } set { this.name = value; } }

		[Browsable(false)]
		public bool NamedInternally { get { return false; } }

		#endregion

		#region Private Members

		private readonly ResourceLoader Resources;

		private uint skinCount;
		private uint vertexCount;
		private uint triangleCount;
		private uint frameCount;
		private uint syncType;
		private uint flags;
		private float size;

		#endregion

		#region Constructor

		/// <summary>Creates an instance of a model and loads it from a stream.</summary>
		/// <param name="s">The stream to load the model data from.</param>
		public Model(ResourceLoader resources, Stream stream) {

			this.Resources = resources;

			BinaryReader ModelReader = new BinaryReader(stream);

			if (Encoding.ASCII.GetString(ModelReader.ReadBytes(4)) != "IDPO") throw new InvalidDataException("Not a valid model file.");
			this.version = ModelReader.ReadUInt32();
			this.scale = new Vector3(stream);
			this.origin = new Vector3(stream);
			this.radius = ModelReader.ReadSingle();
			this.offsets = new Vector3(stream);
			this.skinCount = ModelReader.ReadUInt32();
			this.skinWidth = ModelReader.ReadInt32();
			this.skinHeight = ModelReader.ReadInt32();
			this.vertexCount = ModelReader.ReadUInt32();
			this.triangleCount = ModelReader.ReadUInt32();
			this.frameCount = ModelReader.ReadUInt32();

			if (this.Version >= 6) {
				this.syncType = ModelReader.ReadUInt32();
				this.flags = ModelReader.ReadUInt32();
			}

			this.size = ModelReader.ReadSingle();

			#region Skins

			this.skins = new Skin[this.skinCount];
			for (int i = 0; i < this.skinCount; ++i) {
				this.skins[i] = new Skin(this, stream);
			}

			#endregion

			#region Texture Coordinates

			this.textureCoordinates = new TextureCoordinate[this.vertexCount];
			for (int i = 0; i < this.vertexCount; ++i) {
				this.textureCoordinates[i] = new TextureCoordinate(this, stream);
			}

			#endregion

			#region Triangles

			this.triangles = new Triangle[this.triangleCount];
			for (int i = 0; i < this.triangleCount; ++i) {
				this.triangles[i] = new Triangle(this, stream);
			}

			#endregion

			#region Animation Frames


			this.frames = new AnimationFrame[this.frameCount];
			for (int i = 0; i < this.frameCount; ++i) {
				this.frames[i] = new AnimationFrame();

				// Check if it has multiple frames or not.
				if ((ModelReader.ReadUInt32()) != 0) {

					uint NumSubFrames = ModelReader.ReadUInt32();

					this.frames[i].Min = new Vertex(this, stream);
					this.frames[i].Max = new Vertex(this, stream);


					this.frames[i].Time = new float[NumSubFrames];
					for (int j = 0; j < NumSubFrames; ++j) this.frames[i].Time[j] = ModelReader.ReadSingle();
					this.frames[i].Frames = new AnimationFrame.SimpleFrame[NumSubFrames];
				} else {
					this.frames[i].Frames = new AnimationFrame.SimpleFrame[1];
				}

				for (int j = 0; j < this.frames[i].Frames.Length; ++j) {
					this.frames[i].Frames[j] = new AnimationFrame.SimpleFrame();
					this.frames[i].Frames[j].Min = new Vertex(this, stream);
					this.frames[i].Frames[j].Max = new Vertex(this, stream);
					if (this.Version >= 6) this.frames[i].Frames[j].Name = Encoding.ASCII.GetString(ModelReader.ReadBytes(16)).Split('\0')[0];
					this.frames[i].Frames[j].Vertices = new Vertex[this.vertexCount];
					for (int k = 0; k < this.vertexCount; ++k) {
						this.frames[i].Frames[j].Vertices[k] = new Vertex(this, stream);
					}
				}

			}

			#endregion

		}

		#endregion
	}
}
