using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;

namespace id.Quake.Files {
	public partial class QuakeC {

		/// <summary>Describes a single code statement.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Statement {

			/// <summary>Represents the function of the statement.</summary>
			public enum Operation : ushort {

				/// <summary>Leave the function.</summary>
				OP_DONE,

				/// <summary>C = A * B</summary>
				OP_MUL_F,
				/// <summary>C = A * B (Dot product)</summary>
				OP_MUL_V,
				/// <summary>C = A * B</summary>
				OP_MUL_FV,
				/// <summary>C = A * B</summary>
				OP_MUL_VF,

				/// <summary>C = A / B</summary>
				OP_DIV_F,

				/// <summary>C = A + B</summary>
				OP_ADD_F,
				/// <summary>C = A + B</summary>
				OP_ADD_V,

				/// <summary>C = A - B</summary>
				OP_SUB_F,
				/// <summary>C = A - B</summary>
				OP_SUB_V,

				/// <summary>C = A == B</summary>
				OP_EQ_F,
				/// <summary>C = A == B</summary>
				OP_EQ_V,
				/// <summary>C = A == B</summary>
				OP_EQ_S,
				/// <summary>C = A == B</summary>
				OP_EQ_E,
				/// <summary>C = A == B</summary>
				OP_EQ_FNC,

				/// <summary>C = A != B</summary>
				OP_NE_F,
				/// <summary>C = A != B</summary>
				OP_NE_V,
				/// <summary>C = A != B</summary>
				OP_NE_S,
				/// <summary>C = A != B</summary>
				OP_NE_E,
				/// <summary>C = A != B</summary>
				OP_NE_FNC,

				/// <summary>C = A &lt;= B</summary>
				OP_LE,
				/// <summary>C = A &gt;= B</summary>
				OP_GE,
				/// <summary>C = A &lt; B</summary>
				OP_LT,
				/// <summary>C = A &gt; B</summary>
				OP_GT,

				/// <summary>C = edict A, field B.</summary>
				OP_LOAD_F,
				/// <summary>C = edict A, field B.</summary>
				OP_LOAD_V,
				/// <summary>C = edict A, field B.</summary>
				OP_LOAD_S,
				/// <summary>C = edict A, field B.</summary>
				OP_LOAD_ENT,
				/// <summary>C = edict A, field B.</summary>
				OP_LOAD_FLD,
				/// <summary>C = edict A, field B.</summary>
				OP_LOAD_FNC,
				/// <summary>C = address of edict A, field B.</summary>
				OP_ADDRESS,

				/// <summary>B = A</summary>
				OP_STORE_F,
				/// <summary>B = A</summary>
				OP_STORE_V,
				/// <summary>B = A</summary>
				OP_STORE_S,
				/// <summary>B = A</summary>
				OP_STORE_ENT,
				/// <summary>B = A</summary>
				OP_STORE_FLD,
				/// <summary>B = A</summary>
				OP_STORE_FNC,

				/// <summary>Edict data at offset B = A</summary>
				OP_STOREP_F,
				/// <summary>Edict data at offset B = A</summary>
				OP_STOREP_V,
				/// <summary>Edict data at offset B = A</summary>
				OP_STOREP_S,
				/// <summary>Edict data at offset B = A</summary>
				OP_STOREP_ENT,
				/// <summary>Edict data at offset B = A</summary>
				OP_STOREP_FLD,
				/// <summary>Edict data at offset B = A</summary>
				OP_STOREP_FNC,

				/// <summary>Leave function.</summary>
				OP_RETURN,

				/// <summary>C = A != 0</summary>
				OP_NOT_F,
				/// <summary>C = A != '0 0 0'</summary>
				OP_NOT_V,
				/// <summary>C = A != ""</summary>
				OP_NOT_S,
				/// <summary>C = A != entity 0 (worldspawn).</summary>
				OP_NOT_ENT,
				/// <summary>C = A != function 0.</summary>
				OP_NOT_FNC,

				/// <summary>If A is true then jump B statements.</summary>
				OP_IF,
				/// <summary>If A is false then jump B statements.</summary>
				OP_IFNOT,

				/// <summary>Call A with 0 arguments.</summary>
				OP_CALL0,
				/// <summary>Call A with 1 argument.</summary>
				OP_CALL1,
				/// <summary>Call A with 2 arguments.</summary>
				OP_CALL2,
				/// <summary>Call A with 3 arguments.</summary>
				OP_CALL3,
				/// <summary>Call A with 4 arguments.</summary>
				OP_CALL4,
				/// <summary>Call A with 5 arguments.</summary>
				OP_CALL5,
				/// <summary>Call A with 6 arguments.</summary>
				OP_CALL6,
				/// <summary>Call A with 7 arguments.</summary>
				OP_CALL7,
				/// <summary>Call A with 8 arguments.</summary>
				OP_CALL8,
				
				/// <summary>self.frame = A, self.think = B, self.nextthink = time + 0.1.</summary>
				OP_STATE,

				/// <summary>Jump B statements.</summary>
				OP_GOTO,

				/// <summary>C = A &amp;&amp; B (Logical).</summary>
				OP_AND,
				/// <summary>C = A || B (Logical).</summary>
				OP_OR,

				/// <summary>C = A &amp; B (Bitwise).</summary>
				OP_BITAND,
				/// <summary>C = A | B (Bitwise).</summary>
				OP_BITOR
			}

			private readonly QuakeC Source;

			private readonly Operation operationCode;
			/// <summary>Gets the operation code of this statement.</summary>
			[Description("The operation code of this statement.")]
			public Operation OperationCode {
				get { return this.operationCode; }
			}

			private readonly ushort[] operandIndices;
			/// <summary>Gets the three operand indices (A, B and C).</summary>
			[Description("The indices of the operands.")]
			public ushort[] OperandIndices {
				get { return this.operandIndices; }

			}

			/// <summary>Reads an instance of a statement from a stream.</summary>
			/// <param name="source">The QuakeC source that contains this statement.</param>
			/// <param name="stream">The stream to read the statement from.</param>
			public Statement(QuakeC source, Stream stream) {
				this.Source = source;
				BinaryReader StatementReader = new BinaryReader(stream);
				this.operationCode = (Operation)StatementReader.ReadUInt16();
				if (!Enum.IsDefined(typeof(Operation), this.operationCode)) throw new InvalidDataException();
				this.operandIndices = new ushort[] { 
					StatementReader.ReadUInt16(),
					StatementReader.ReadUInt16(),
					StatementReader.ReadUInt16() 
				};
			}
		}
	}
}