using System;
using System.IO;
using System.Collections.Generic;

namespace id.Quake {

	/// <summary>A class that can be used to load resources from files from the native file system and archives.</summary>
	public class ResourceLoader {

		#region Properties

		private readonly string baseDirectory;
		/// <summary>Gets the base directory used to load files.</summary>
		public string BaseDirectory {
			get { return this.baseDirectory;}
		}

		private Dictionary<string, Packages.IPackage> FilesInPackages;
		private List<Packages.IPackage> AllPackages;

		private Dictionary<string, object> Cache;

		#endregion

		#region Constructor

		/// <summary>Creates an instance of the resource loader.</summary>
		/// <param name="baseDirectory">Specify the base game directory to load files from.</param>
		public ResourceLoader(string baseDirectory) {
			this.baseDirectory = baseDirectory;
			this.FilesInPackages = new Dictionary<string, id.Quake.Packages.IPackage>();
			this.AllPackages = new List<id.Quake.Packages.IPackage>();
			this.Cache = new Dictionary<string, object>();
		}
		
		#endregion

		#region Public Methods

		/// <summary>Open a file.</summary>
		/// <param name="filename">The relative path of the filename from the base directory.</param>
		/// <returns>A stream that can be used to decode the file.</returns>
		public Stream OpenFile(string filename) {
			string Filename = FixFilename(filename);
			Packages.IPackage FromPackage;

			/*if (File.Exists(Path.Combine(this.BaseDirectory, filename))) {
				return File.OpenRead(Path.Combine(this.BaseDirectory, filename));
			} else {
				if (this.FilesInPackages.TryGetValue(Filename, out FromPackage)) {
					return FromPackage.OpenFile(filename);
				} else {
					throw new FileNotFoundException();
				}
			}*/

			if (this.FilesInPackages.TryGetValue(Filename, out FromPackage)) {
				return FromPackage.OpenFile(filename);
			} else {
				return File.OpenRead(Path.Combine(this.BaseDirectory, filename));
			}
		}


		/// <summary>Add a package file to the loader.</summary>
		/// <param name="package">The package to extract files from.</param>
		public void AddPackage(Packages.IPackage package) {
			if (this.AllPackages.Contains(package)) return;
			this.AllPackages.Add(package);
			foreach (string s in package.GetFiles()) {
				string Filename = FixFilename(s);
				if (FilesInPackages.ContainsKey(Filename)) FilesInPackages.Remove(Filename);
				this.FilesInPackages.Add(Filename, package);
			}
		}

		public void AddStandardPackages() {
			string[] Packages = Directory.GetFiles(this.BaseDirectory, "PAK*.PAK");
			Array.Sort<string>(Packages);
			Array.Reverse(Packages);
			foreach (string s in Packages) AddPackage(new Packages.Pack(File.OpenRead(s)));
		}

		/// <summary>Get a list of all of the available files.</summary>
		public string[] GetFiles() {
			List<string> AllFileNames = new List<string>(256);

			foreach (Packages.IPackage pack in this.AllPackages) {
				foreach (string s in pack.GetFiles()) {
					string t = FixFilename(s);
					if (!AllFileNames.Contains(t)) AllFileNames.Add(t);
				}
			}

			foreach (string s in Directory.GetFiles(this.BaseDirectory, "*", SearchOption.AllDirectories)) {
				string t = FixFilename(Path.GetFullPath(s).Substring(this.BaseDirectory.Length + 1));
				if (!AllFileNames.Contains(t)) AllFileNames.Add(t);
			}

			return AllFileNames.ToArray();

		}

		/// <summary>Add an item to the cache for frequently used resources.</summary>
		/// <param name="cachedName">Name of the item to cache.</param>
		/// <param name="objectToCache">The item to cache.</param>
		public void CacheObject(string cachedName, object objectToCache) {
			if (this.Cache.ContainsKey(cachedName)) this.Cache.Remove(cachedName);
			this.Cache.Add(cachedName, objectToCache);
		}

		/// <summary>Try and extract an item from the cache.</summary>
		/// <param name="cachedName">Name of the item to retrieve.</param>
		/// <param name="cachedObject">Outputs the item if found.</param>
		public bool TryGetCachedObject<T>(string cachedName, out T cachedObject) {
			object o; bool b;
			b = this.Cache.TryGetValue(cachedName, out o);
			if (b) {
				cachedObject = (T)o;
			} else {
				cachedObject = default(T);
			}
			return b;
		}

		/// <summary>Check to see if a particular file exists.</summary>
		/// <param name="name">Name of the file to check for.</param>
		/// <returns>True if the file exists, false otherwise.</returns>
		public bool FileExists(string name) {
			if (File.Exists(Path.Combine(this.BaseDirectory, name))) return true;
			foreach (Packages.IPackage pack in this.AllPackages) {
				if (pack.FileExists(name)) return true;
			}
			return false;
		}

		/// <summary>Check if a file comes from a package file or from the local file system.</summary>
		/// <param name="name">The name of the file to check.</param>
		/// <returns>True if the file comes from a package, false if it's from the local file system.</returns>
		public bool FileIsPackaged(string name) {
			return this.FilesInPackages.ContainsKey(FixFilename(name));
		}


		/// <summary>Load a complete object from a filename.</summary>
		/// <typeparam name="T">The type of the object to load.</typeparam>
		/// <param name="filename">The name of the object to load.</param>
		public T LoadObject<T>(string filename) where T : INamedResource {
			T Result = default(T);
			using (Stream S = this.OpenFile(filename)) {
				try {
					Result = (T)typeof(T).GetConstructor(new Type[] { typeof(ResourceLoader), typeof(Stream) }).Invoke(new object[] { this, S });
					if (!Result.NamedInternally) Result.Name = FixFilename(filename);
				} catch (Exception e) {
					throw e.InnerException;
				}
			}
			return Result;
		}

		/// <summary>Amend a filename into the expected format.</summary>
		/// <param name="name">The filename to fix.</param>
		/// <remarks>Internally filenames need to be entirely in lowercase, with forward slashes being used as path delimiters.</remarks>
		public static string FixFilename(string name) {
			return name.ToLowerInvariant().Replace('\\', '/');
		}

		#endregion



	}


}