using System;
using System.ComponentModel;
using System.Drawing.Design;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using id.Quake.Files;

namespace id.Quake {

	/// <summary>Defines a 2D bitmap in a standardised 32bpp ARGB format.</summary>
	[TypeConverter(typeof(ExpandableObjectConverter))]
	[Editor(typeof(PictureEditor), typeof(UITypeEditor))]
	public class Picture {

		#region Properties

		private readonly int width;
		/// <summary>Gets the width of the picture in pixels.</summary>
		public int Width { get { return this.width; } }

		private readonly int height;
		/// <summary>Gets the height of the picture in pixels.</summary>
		public int Height { get { return this.height; } }

		private readonly int[] data;
		/// <summary>Gets the picture data as an array of 32bpp ARGB pixels</summary>
		public int[] Data {
			get { return this.data; }
		}

		public override string ToString() {
			return string.Format("{0}x{1}", this.width, this.height);
		}

		#endregion

		#region Methods

		/// <summary>Returns the data padded to a particular stride.</summary>
		/// <param name="stride">The stride of a scanline in bytes (must be a multiple of four).</param>
		/// <returns>The data with scanlines padded to a particular stride width.</returns>
		public int[] GetPaddedData(int stride) {
			if ((stride & 3) != 0) throw new ArgumentException("Stride must be a multiple of four.");
			int[] Result = new int[(stride / 4) * this.Height];

			int Source = 0, Destination = 0, StrideWidth = stride / 4;

			for (int i = 0; i < this.Height; ++i) {
				Array.Copy(this.data, Source, Result, Destination, this.Width);
				Source += this.Width;
				Destination += StrideWidth;
			}

			return Result;
		}

		/// <summary>Gets a copy of the picture resized to a particular width and height (nearest-neighbour).</summary>
		/// <param name="width">Width of the destination picture.</param>
		/// <param name="height">Height of the destination picture.</param>
		public Picture GetResized(int width, int height, bool smoothed) {
			if (width == this.width && height == this.height) return this;
			int[] Resized = new int[width * height];


			for (int Destination = 0, y = 0; y < height; ++y) {
				for (int Row = ((y * this.height) / height) * this.width, x = 0; x < width; ++x) {

					int Column = (x * this.width) / width;

					if (smoothed) {

						int Column2 = Math.Min(Column + 1, this.width - 1);
						int Row2 = Math.Min(Row + this.width, this.width * (this.height - 1));

						int a = this.data[Row + Column];
						int b = this.data[Row + Column2];
						int c = this.data[Row2 + Column];
						int d = this.data[Row2 + Column2];
						Resized[Destination++] = ColourInterpolate2D(a, b, c, d, (byte)((x * 256) / this.width), (byte)((y * 256) / this.height));



					} else {
						Resized[Destination++] = this.data[Row + Column];
					}

				}
			}

			return new Picture(width, height, Resized);
		}


		public Picture GetCropped(int x, int y, int width, int height) {
			int[] Cropped = new int[width * height];

			int Source = y * this.width + x;
			int Dest = 0;

			for (int cy = 0; cy < height; ++cy) {
				for (int cx = 0; cx < width; ++cx) {
					Cropped[Dest++] = this.data[Source++];
				}
				Source += this.width - width;
			}
			return new Picture(width, height, Cropped);
		}

		struct Colour {
			public int A, R, G, B;

			public static Colour FromArgb(int argb) {
				Colour C = new Colour();
				C.B = (argb & 0xFF); argb >>= 8;
				C.G = (argb & 0xFF); argb >>= 8;
				C.R = (argb & 0xFF); argb >>= 8;
				C.A = (argb & 0xFF);
				return C;
			}
			public Colour(int a, int r, int g, int b) {
				this.A = a; this.R = r; this.G = g; this.B = b;
			}
			public Colour(int all) {
				this.A = all; this.R = all; this.G = all; this.B = all;
			}

			public static Colour operator +(Colour a, Colour b) { return new Colour(a.A + b.A, a.R + b.R, a.G + b.G, a.B + b.B); }
			public static Colour operator +(Colour a, int b) { return a + new Colour(b); }
			public static Colour operator +(int a, Colour b) { return new Colour(a) + b; }
			public static Colour operator -(Colour a, Colour b) { return new Colour(a.A - b.A, a.R - b.R, a.G - b.G, a.B - b.B); }
			public static Colour operator -(Colour a, int b) { return a - new Colour(b); }
			public static Colour operator -(int a, Colour b) { return new Colour(a) - b; }
			public static Colour operator /(Colour a, Colour b) { return new Colour(a.A / b.A, a.R / b.R, a.G / b.G, a.B / b.B); }
			public static Colour operator /(Colour a, int b) { return a / new Colour(b); }
			public static Colour operator /(int a, Colour b) { return new Colour(a) / b; }
			public static Colour operator *(Colour a, Colour b) { return new Colour(a.A * b.A, a.R * b.R, a.G * b.G, a.B * b.B); }
			public static Colour operator *(Colour a, int b) { return a * new Colour(b); }
			public static Colour operator *(int a, Colour b) { return new Colour(a) * b; }

			public static Colour Interpolate(Colour a, Colour b, byte amount) {
				return a + ((b - a) * amount) / 256;
			}
			public int ToArgb() {
				return (this.A << 24) | (this.R << 16) | (this.G << 8) | this.B;
			}
		}

		private int ColourInterpolate2D(int a, int b, int c, int d, byte x, byte y) {
			return Colour.Interpolate(Colour.Interpolate(Colour.FromArgb(a), Colour.FromArgb(b), x), Colour.Interpolate(Colour.FromArgb(c), Colour.FromArgb(d), x), y).ToArgb();
		}

		/// <summary>Gets a copy of the picture resized to a particular width and height.</summary>
		/// <param name="width">Width of the destination picture.</param>
		/// <param name="height">Height of the destination picture.</param>
		/// <param name="smoothed">True to smoothly scale up the picture, false to use nearest-neighbour.</param>
		public Picture GetResized(int width, int height) {
			return this.GetResized(width, height, false);
		}


		/// <summary>Gets a copy of the picture scaled up so that both width and height are powers of two.</summary>
		/// <param name="smoothed">True to smoothly scale up the picture, false to use nearest-neighbour.</param>
		public Picture GetNearestPowerOfTwo(bool smoothed) {
			int PowWidth = 1; while (PowWidth < this.Width) PowWidth <<= 1;
			int PowHeight = 1; while (PowHeight < this.Height) PowHeight <<= 1;
			return this.GetResized(PowWidth, PowHeight, smoothed);
		}

		/// <summary>Gets a copy of the picture scaled up so that both width and height are powers of two (nearest-neighbour).</summary>
		public Picture GetNearestPowerOfTwo() {
			return this.GetNearestPowerOfTwo(false);
		}

		#endregion

		#region Constructor

		/// <summary>Creates an instance of the picture class.</summary>
		/// <param name="width">Width of the picture in pixels.</param>
		/// <param name="height">Height of the picture in pixels.</param>
		/// <param name="data">Array of ints of pixel data, width * height in length.</param>
		public Picture(int width, int height, int[] data) {
			this.width = width;
			this.height = height;
			this.data = data;
		}

		#endregion

	}

	public class PictureEditor : UITypeEditor {

		public override bool GetPaintValueSupported(System.ComponentModel.ITypeDescriptorContext context) {
			return true;
		}

		public override void PaintValue(PaintValueEventArgs e) {
			IPicture IP = e.Context.Instance as IPicture;
			if (IP != null) {
				Picture P = IP.Picture;
				if (P != null) {
					using (Bitmap B = new Bitmap(P.Width, P.Height, PixelFormat.Format32bppRgb)) {
						BitmapData BD = B.LockBits(new Rectangle(0, 0, P.Width, P.Height), ImageLockMode.WriteOnly, PixelFormat.Format32bppRgb);
						int[] SourceData = P.GetPaddedData(BD.Stride);
						Marshal.Copy(SourceData, 0, BD.Scan0, SourceData.Length);
						B.UnlockBits(BD);
						e.Graphics.DrawImage(B, e.Bounds);
					}
				}
			}
		}

	}

}