using System;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.IO;

namespace id.Quake {

	/// <summary>Defines a vector with three components.</summary>
	[StructLayout(LayoutKind.Explicit)]
	public struct Vector3  {

		#region Properties

		/// <summary>Gets or sets the X component of the vector.</summary>
		[FieldOffset(0)]
		public float X;

		/// <summary>Gets or sets the Y component of the vector.</summary>
		[FieldOffset(4)]
		public float Y;

		/// <summary>Gets or sets the Z component of the vector.</summary>
		[FieldOffset(8)]
		public float Z;

		/// <summary>Gets the magnitude of the vector.</summary>
		public float Magnitude {
			get { return (float)Math.Sqrt(X * X + Y * Y + Z * Z); }
		}

		/// <summary>Gets the magnitude squared of the vector.</summary>
		public float MagnitudeSquared {
			get { return X * X + Y * Y + Z * Z; }
		}

		#endregion

		#region Constructors

		/// <summary>Initialise a Vector3 setting all components to the same value.</summary>
		/// <param name="value">Value to initialise each component to.</param>
		public Vector3(float value) {
			this.X = value; this.Y = value; this.Z = value;
		}

		/// <summary>Initialise a Vector3 specifing values for all three component values.</summary>
		/// <param name="x">Initial value for the x-component of the vector.</param>
		/// <param name="y">Initial value for the y-component of the vector.</param>
		/// <param name="z">Initial value for the z-component of the vector.</param>
		public Vector3(float x, float y, float z) {
			this.X = x; this.Y = y; this.Z = z;
		}

		/// <summary>Initialise a Vector3 reading its component values from a stream.</summary>
		public Vector3(System.IO.Stream stream) {
			System.IO.BinaryReader VectorReader = new System.IO.BinaryReader(stream);
			this.X = VectorReader.ReadSingle(); this.Y = VectorReader.ReadSingle(); this.Z = VectorReader.ReadSingle();
		}

		/// <summary>Parse a Vector3 from a string.</summary>
		/// <param name="s">The string to parse.</param>
		public static Vector3 Parse(string s) {
			string[] Components = s.Split(' ');
			if (Components.Length != 3) throw new InvalidDataException();
			return new Vector3(float.Parse(Components[0]), float.Parse(Components[1]), float.Parse(Components[2]));
		}

		#endregion

		#region Methods

		/// <summary>Calculates the dot (scalar) product of two vectors.</summary>
		public static float Dot(Vector3 a, Vector3 b) {
			return a.X * b.X + a.Y * b.Y + a.Z * b.Z;
		}

		/// <summary>Calculates the cross (vector) product of two vectors.</summary>
		public static Vector3 Cross(Vector3 a, Vector3 b) {
			return new Vector3(a.Y * b.Z - b.Y * a.Z, a.Z * b.X - b.Z * a.X, a.X * b.Y - b.X * a.Y);
		}


		/// <summary>Returns a vector made up of the smallest components of two vectors.</summary>
		public static Vector3 Min(Vector3 a, Vector3 b) {
			return new Vector3(Math.Min(a.X, b.X), Math.Min(a.Y, b.Y), Math.Min(a.Z, b.Z));
		}


		/// <summary>Returns a vector made up of the largest components of two vectors.</summary>
		public static Vector3 Max(Vector3 a, Vector3 b) {
			return new Vector3(Math.Max(a.X, b.X), Math.Max(a.Y, b.Y), Math.Max(a.Z, b.Z));
		}

		public static Vector3 ReflectByNormal(Vector3 direction, Vector3 normal) {
			return -direction;
		}

		/// <summary>Changes the coefficients of the vector to make it of unit length.</summary>
		public void Normalise() {
			float InverseMagnitude = 1f / this.Magnitude;
			this.X *= InverseMagnitude;
			this.Y *= InverseMagnitude;
			this.Z *= InverseMagnitude;
		}

		/// <summary>Calculate vectors that point forwards, right and up based on a pitch, yaw and roll rotation.</summary>
		/// <param name="angles">The pitch (X), yaw (Y) and roll (Z) angles in radians.</param>
		/// <param name="forward">Outputs a vector pointing forwards based on the input angles.</param>
		/// <param name="right">Outputs a vector pointing right based on the input angles.</param>
		/// <param name="up">Outputs a vector pointing up based on the input angles.</param>
		public static void GetAngleVectors(Vector3 angles, out Vector3 forward, out Vector3 right, out Vector3 up) {
			// XYZ = Pitch Yaw Roll
			Angle Angle;
			float SinRoll, SinPitch, SinYaw, CosRoll, CosPitch, CosYaw;

			Angle = new Angle(angles.X);
			SinPitch = (float)Math.Sin(Angle.Radians);
			CosPitch = (float)Math.Cos(Angle.Radians);

			Angle = new Angle(angles.Y);
			SinYaw = (float)Math.Sin(Angle.Radians);
			CosYaw = (float)Math.Cos(Angle.Radians);

			Angle = new Angle(angles.Z);
			SinRoll = (float)Math.Sin(Angle.Radians);
			CosRoll = (float)Math.Cos(Angle.Radians);

			forward = new Vector3(CosPitch * CosYaw, CosPitch * SinYaw, -SinPitch);
			right = new Vector3(-1 * SinRoll * SinPitch * CosYaw + -1 * CosRoll * -SinYaw, -1 * SinRoll * SinPitch * SinYaw + -1 * CosRoll * CosYaw, -1 * SinRoll * CosPitch);
			up = new Vector3(CosRoll * SinPitch * CosYaw + -SinRoll * -SinYaw, CosRoll * SinPitch * SinYaw + -SinRoll * CosYaw, CosRoll * CosPitch);
		}

		/// <summary>Reflect a vector against a plane's surface normal.</summary>
		/// <param name="incident">The incident vector to reflect.</param>
		/// <param name="normal">The normal of the plane to reflect against.</param>
		public static Vector3 Reflect(Vector3 incident, Vector3 normal) {
			return incident - 2 * (Vector3.Dot(incident, normal)) * normal;
		}

		#endregion

		#region Operators

		/// <summary>Adds the components of two vectors together.</summary>
		public static Vector3 operator +(Vector3 a, Vector3 b) { return new Vector3(a.X + b.X, a.Y + b.Y, a.Z + b.Z); }

		/// <summary>Adds a constant value to the components of a vector.</summary>
		public static Vector3 operator +(Vector3 a, float b) { return a + new Vector3(b); }

		/// <summary>Adds a constant value to the components of a vector.</summary>
		public static Vector3 operator +(float a, Vector3 b) { return new Vector3(a) + b; }

		/// <summary>Subtracts the components of one vector from another.</summary>
		public static Vector3 operator -(Vector3 a, Vector3 b) { return new Vector3(a.X - b.X, a.Y - b.Y, a.Z - b.Z); }

		/// <summary>Subtracts a constant from the components of a vector.</summary>
		public static Vector3 operator -(Vector3 a, float b) { return a - new Vector3(b); }

		/// <summary>Subtracts a constant from the components of a vector.</summary>
		public static Vector3 operator -(float a, Vector3 b) { return new Vector3(a) - b; }

		/// <summary>Negates the components of a vector.</summary>
		public static Vector3 operator -(Vector3 a) { return new Vector3(-a.X, -a.Y, -a.Z); }

		/// <summary>Multiplies the components of one vector with another.</summary>
		public static Vector3 operator *(Vector3 a, Vector3 b) { return new Vector3(a.X * b.X, a.Y * b.Y, a.Z * b.Z); }

		/// <summary>Multiplies the components of a vector with a constant value.</summary>
		public static Vector3 operator *(Vector3 a, float b) { return a * new Vector3(b); }

		/// <summary>Multiplies the components of a vector with a constant value.</summary>
		public static Vector3 operator *(float a, Vector3 b) { return new Vector3(a) * b; }

		/// <summary>Divides the components of one vector by the components of another.</summary>
		public static Vector3 operator /(Vector3 a, Vector3 b) { return new Vector3(a.X / b.X, a.Y / b.Y, a.Z / b.Z); }

		/// <summary>Divides the components of a vector by a constant value.</summary>
		public static Vector3 operator /(Vector3 a, float b) { return a / new Vector3(b); }

		/// <summary>Divides a vector with all components set to a single value by another vector.</summary>
		public static Vector3 operator /(float a, Vector3 b) { return new Vector3(a) / b; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have the same components, false otherwise.</returns>
		public static bool operator ==(Vector3 a, Vector3 b) { return a.X == b.X && a.Y == b.Y && a.Z == b.Z; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have differing components, false if both are the same.</returns>
		public static bool operator !=(Vector3 a, Vector3 b) { return a.X != b.X || a.Y != b.Y || a.Z != b.Z; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have the same components, false otherwise.</returns>
		public override bool Equals(object obj) {
			if (obj.GetType() != this.GetType()) return false;
			return (Vector3)obj == this;
		}

		#endregion

		/// <summary>Returns a formatted string describing the vector's components.</summary>
		public override string ToString() {
			return string.Format("'{0:N1} {1:N1} {2:N1}'", this.X, this.Y, this.Z);
		}

		/// <summary>Gets the hash code for this vector.</summary>
		public override int GetHashCode() {
			return this.X.GetHashCode() ^ this.Y.GetHashCode() ^ this.Z.GetHashCode();
		}

		#region Static Properties

		/// <summary>Gets a Vector3 with all components set to zero.</summary>
		public static Vector3 Zero {
			get { return new Vector3(0f); }
		}

		#endregion


	}


}