using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace id.Quake {
	public partial class VirtualMachine {

		#region Types


		/// <summary>Defines a console variable.</summary>
		public class ConsoleVariable {

			#region Storage

			/// <summary>The stored value (can be a float or a string).</summary>
			private object Value = 0f;
			/// <summary>True if set from a string value, false if set from a float.</summary>
			private bool SetFromString;

			/// <summary>True if archive and notifyPlayers set on purpose, false if unknown.</summary>
			internal bool KnowsFlags = false;

			#endregion

			#region Properties

			private readonly string name;
			/// <summary>Gets the name of the console variable.</summary>
			public string Name { get { return this.name; } }

			/// <summary>Gets or sets the value of the variable as a string.</summary>
			public string String {
				get {
					return SetFromString ? (string)Value : ((float)Value).ToString();
				}
				set {
					this.Value = value;
					this.SetFromString = true;
				}
			}

			private readonly bool archive;
			/// <summary>Returns true if this variable needs to be saved to the configuration file.</summary>
			public bool Archive {
				get { return this.archive; }
			}

			private readonly bool notifyPlayers;
			/// <summary>Returns true if players are notified when the variable has been changed.</summary>
			public bool NotifyPlayers {
				get { return this.notifyPlayers; }
			}

			/// <summary>Gets or sets the variable as a float.</summary>
			public float Float {
				get {
					return SetFromString ? float.Parse((string)this.Value, System.Globalization.CultureInfo.InvariantCulture) : (float)this.Value;
				}
				set {
					this.Value = value;
					this.SetFromString = false;
				}
			}

			/// <summary>Returns a formatted string describing the variable.</summary>
			public override string ToString() {
				return this.name + "\t\"" + this.String + "\"";
			}

			#endregion

			#region Constructors

			/// <summary>Create an instance of console variable class.</summary>
			public ConsoleVariable(string name) {
				this.name = name;
			}

			/// <summary>Create an instance of console variable class, setting its value to a float.</summary>
			/// <param name="name">The name of the console variable.</param>
			/// <param name="value">The value to initialise the console variable to.</param>
			public ConsoleVariable(string name, float value)
				: this(name) {
				this.Float = value;
			}

			/// <summary>Create an instance of console variable class, setting its value to a string.</summary>
			/// <param name="name">The name of the console variable.</param>
			/// <param name="value">The value to initialise the console variable to.</param>
			public ConsoleVariable(string name, string value)
				: this(name) {
				this.String = value;
			}

			/// <summary>Create an instance of console variable class, setting its value to a float.</summary>
			/// <param name="name">The name of the console variable.</param>
			/// <param name="value">The value to initialise the console variable to.</param>
			/// <param name="archive">Set to true if the variable needs to be saved to the configuration file.</param>
			public ConsoleVariable(string name, float value, bool archive)
				: this(name, value) {
				this.archive = archive;
			}

			/// <summary>Create an instance of console variable class, setting its value to a string.</summary>
			/// <param name="name">The name of the console variable.</param>
			/// <param name="value">The value to initialise the console variable to.</param>
			/// <param name="archive">Set to true if the variable needs to be saved to the configuration file.</param>
			public ConsoleVariable(string name, string value, bool archive)
				: this(name, value) {
				this.archive = archive;
			}


			/// <summary>Create an instance of console variable class, setting its value to a float.</summary>
			/// <param name="name">The name of the console variable.</param>
			/// <param name="value">The value to initialise the console variable to.</param>
			/// <param name="archive">Set to true if the variable needs to be saved to the configuration file.</param>
			/// <param name="notifyPlayers">Set to true if players are notified when the variable is changed.</param>
			public ConsoleVariable(string name, float value, bool archive, bool notifyPlayers)
				: this(name, value, archive) {
				this.notifyPlayers = notifyPlayers;
			}

			/// <summary>Create an instance of console variable class, setting its value to a string.</summary>
			/// <param name="name">The name of the console variable.</param>
			/// <param name="value">The value to initialise the console variable to.</param>
			/// <param name="archive">Set to true if the variable needs to be saved to the configuration file.</param>
			/// 			/// <param name="notifyPlayers">Set to true if players are notified when the variable is changed.</param>
			public ConsoleVariable(string name, string value, bool archive, bool notifyPlayers)
				: this(name, value, archive) {
				this.notifyPlayers = notifyPlayers;
			}

			#endregion

			#region Parsing

			/// <summary>Parse a console variable.</summary>
			/// <param name="cvar">The string to parse.</param>
			/// <param name="value">The console variable to return.</param>
			/// <returns>True on success, false on failure.</returns>
			/// <remarks>This will handle the special syntax in the <c>cvars.dat</c> file as well as the more traditional <c>key "value"</c> syntax as used in <c>cfg</c> files.</remarks>
			public static bool TryParse(string cvar, out ConsoleVariable value) {

				value = null;

				// Remove comments and whitespace.
				cvar = cvar.Split(new string[] { "//" }, StringSplitOptions.None)[0].Trim();

				// Split source line into components.
				List<string> Components = new List<string>();
				List<char> TokenSeperators = new List<char>(new char[] { ' ', '\t', ',', '}', '{' });

				for (int i = 0; i < cvar.Length; ++i) {
					char c = cvar[i];
					if (c == ' ' || c == '\t') continue;
					string Component = "";
					if (c == '"') {
						for (++i; i < cvar.Length && cvar[i] != '"'; ++i) {
							Component += cvar[i];
						}
					} else {
						for (; i < cvar.Length && !TokenSeperators.Contains(cvar[i]); ++i) {
							Component += cvar[i];
						}
					}
					if (Component != "") Components.Add(Component);
				}

				if (Components.Count < 2) return false;

				string Name = null;
				string Value = null;
				bool Archive = false;
				bool NotifyPlayer = false;
				bool KnowsFlags = false;

				if (Components.Count == 2) {
					// Simple CVAR in "key value" format.
					Name = Components[0];
					Value = Components[1];
				} else {
					// Format will be 'name' '=' 'name' 'value' ['save'] ['notifies player when changed'] ';'
					if (Components.Count < 5 || Components.Count > 7) return false;
					if (Components[Components.Count - 1] != ";") return false;
					if (Components[1] != "=") return false;
					if (Components[0] != Components[2]) return false;
					Name = Components[2];
					Value = Components[3];
					if (Components.Count >= 6) Archive = Components[4].ToLowerInvariant() != "false";
					if (Components.Count >= 7) NotifyPlayer = Components[5].ToLowerInvariant() != "false";
					KnowsFlags = true;
				}


				float CheckFloat;
				bool IsFloat = float.TryParse(Value, out CheckFloat);

				if (IsFloat) {
					value = new ConsoleVariable(Name, CheckFloat, Archive, NotifyPlayer);
				} else {
					value = new ConsoleVariable(Name, Value, Archive, NotifyPlayer);
				}
				value.KnowsFlags = KnowsFlags;
				return true;
			}


			/// <summary>Parse a console variable.</summary>
			/// <param name="cvar">The string to parse.</param>
			/// <remarks>This will handle the special syntax in the <c>cvars.dat</c> file as well as the more traditional <c>key "value"</c> syntax as used in <c>cfg</c> files.</remarks>
			public static ConsoleVariable Parse(string cvar) {
				ConsoleVariable Value;
				if (TryParse(cvar, out Value)) {
					return Value;
				} else {
					throw new FormatException("Could not parse the console variable.");
				}
			}

			#endregion

			#region Methods

			/// <summary>Copy a value (and value only) from another variable.</summary>
			/// <param name="source">The source variable to load from.</param>
			internal void CopyValue(ConsoleVariable source) {
				this.Value = source.Value;
				this.SetFromString = source.SetFromString;
			}

			#endregion


		}

		/// <summary>Defines a collection of console variables.</summary>
		public class ConsoleVariableCollection {

			private Dictionary<string, ConsoleVariable> Variables;
			private static string FixName(string name) {
				return name.Trim().ToLowerInvariant();
			}

			/// <summary>Gets a console variable by its name.</summary>
			/// <param name="name">The name of the console variable to access.</param>
			public ConsoleVariable this[string name] {
				get {
					ConsoleVariable Value;
					if (this.Variables.TryGetValue(FixName(name), out Value)) return Value;
					throw new KeyNotFoundException("Console variable not set.");
				}
			}

			/// <summary>Try and get a console variable from a string name.</summary>
			/// <param name="name">The name of the variable to search for.</param>
			/// <param name="value">The variable to store the result in if found.</param>
			/// <returns>Trye if the variable was found, false otherwise.</returns>
			public bool TryGetValue(string name, out ConsoleVariable value) {
				return this.Variables.TryGetValue(FixName(name), out value);				
			}

			/// <summary>Get or create a variable.</summary>
			/// <param name="name">Name of the variable to access.</param>
			private ConsoleVariable GetOrCreateVariable(string name) {
				ConsoleVariable Value;
				if (!this.Variables.TryGetValue(name, out Value)) {
					Value = new ConsoleVariable(name, 0f);
					this.Variables.Add(name, Value);
				}
				return Value;

			}

			/// <summary>Set a console variable to a particular float value.</summary>
			/// <param name="name">The name of the console variable to change.</param>
			/// <param name="value">The new value of the console variable.</param>
			public void SetVariable(string name, float value) {
				this.GetOrCreateVariable(name).Float = value;
			}

			/// <summary>Set a console variable to a particular string value.</summary>
			/// <param name="name">The name of the console variable to change.</param>
			/// <param name="value">The new value of the console variable.</param>
			public void SetVariable(string name, string value) {
				this.GetOrCreateVariable(name).String = value;
			}

			/// <summary>Creates an instance of the ConsoleVariableCollection class.</summary>
			/// <param name="resources">A ResourceLoader to use to load the <c>cvars.dat</c> file from.</param>
			public ConsoleVariableCollection(ResourceLoader resources) {
				this.Variables = new Dictionary<string, ConsoleVariable>();

				if (resources.FileExists("cvars.dat")) {
					using (TextReader CVarReader = new StreamReader(resources.OpenFile("cvars.dat"))) {
						string Line = "";
						while ((Line = CVarReader.ReadLine()) != null) {
							this.LoadVariable(Line);
						}
					}
				}


			}

			/// <summary>Load a variable from a string.</summary>
			/// <param name="cvar">The console variable declaration string.</param>
			public void LoadVariable(string cvar) {
				// Parse the variable:
				ConsoleVariable CV;
				if (!ConsoleVariable.TryParse(cvar, out CV)) return;

				// Try and find an existing variable. If we didn't know flags, and the new one does, remove it.
				ConsoleVariable Existing;
				if (this.Variables.TryGetValue(CV.Name, out Existing)) {
					if (!Existing.KnowsFlags && CV.KnowsFlags) {
						this.Variables.Remove(Existing.Name);
					}
				}

				if (!this.Variables.ContainsKey(CV.Name)) {
					this.Variables.Add(CV.Name, CV);
				} else {
					this.Variables[CV.Name].CopyValue(CV);
				}
			}

		}
		
		#endregion

		private readonly ConsoleVariableCollection consoleVariables;
		/// <summary>Gets the console variable collection.</summary>
		public ConsoleVariableCollection ConsoleVariables {
			get { return this.consoleVariables; }
		}


	}
}