using System;
using System.Collections.Generic;
using id.Quake.Files;

namespace id.Quake {

	/// <summary>Provides a VirtualMachine to run the game on.</summary>
	/// <remarks>
	/// The virtual machine manages memory for edicts (entities) internally.
	/// However, changes made to global variables are made within the QuakeC program.
	/// </remarks>
	public partial class VirtualMachine {



		#region Private Members

		/// <summary>Current statement pointer.</summary>
		private int StatementPointer;


		/// <summary>Resource loader used to load cached files.</summary>
		private readonly ResourceLoader Resources;

		#endregion

		#region Properties

		private readonly QuakeC quakeCSource;
		/// <summary>Gets the QuakeC source file used by the virtual machine.</summary>
		public QuakeC QuakeCSource {
			get { return this.quakeCSource; }
		}

		private EdictCollection edicts;
		/// <summary>Gets the collection of edicts used by the virtual machine.</summary>
		public EdictCollection Edicts {
			get { return this.edicts; }
		}

		private Level level;
		/// <summary>Gets the BSP level used by the virtual machine.</summary>
		public Level Level {
			get { return this.level; }
		}

		#endregion

		#region Constructor

		/// <summary>Creates an instance of the VirtualMachine.</summary>
		/// <param name="resources">A <c>ResourceLoader</c> that can be used to load resources as required by the virtual machine.</param>
		/// <param name="source">The compiled QuakeC program to run.</param>
		public VirtualMachine(ResourceLoader resources, QuakeC source) {

			this.Resources = resources;
			this.quakeCSource = source;
			this.edicts = new EdictCollection(this.quakeCSource);
			this.level = null;

			this.Self = this.QuakeCSource.Globals["self"];
			this.Other = this.QuakeCSource.Globals["other"];
			this.Time = this.QuakeCSource.Globals["time"];


			this.StatementPointer = 0;

			this.InitialiseBuiltInFunctions();
			this.InitialiseFunctions();

			#region Working Variables

			this.Return = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Return);
			this.Parameter0 = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Parameter0);
			this.Parameter1 = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Parameter1);
			this.Parameter2 = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Parameter2);
			this.Parameter3 = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Parameter3);
			this.Parameter4 = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Parameter4);
			this.Parameter5 = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Parameter5);
			this.Parameter6 = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Parameter6);
			this.Parameter7 = new WorkingVariable(this, (ushort)WorkingVariableOffsets.Parameter7);

			#endregion

			this.consoleVariables = new ConsoleVariableCollection(this.Resources);

		}

		#endregion

		#region Methods

		/// <summary>Load a level into the virtual machine.</summary>
		/// <param name="level">The level to load.</param>
		/// <remarks>The levels edicts will be loaded.</remarks>
		public void LoadLevel(Level level) {
			this.level = level;
						
			// Clear out the model cache:
			this.ModelCache = new List<object>();
			this.ModelCacheIndex = new Dictionary<string, int>();

			//HACK: Is this correct?
			for (int i = 0; i < level.Models.Length; i++) {
				this.CacheModel("*" + i);
			}
			

			this.edicts.LoadFromLevel(level, this);
		}

		#endregion

	}
}
