#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using System.IO;
using Quake = id.Quake;
using System.Diagnostics;
using System.Reflection;
using System.Threading;
using VectorRendererComponent;
#endregion

namespace XnaQuake {
	/// <summary>
	/// This is the main type for your game
	/// </summary>
	public class QuakeGame : Microsoft.Xna.Framework.Game {


		private VectorRenderer VR;
		private GraphicsDeviceManager Graphics;
		private ContentManager Content;

		private Quake.ResourceLoader Resources;
		private Renderer.RenderController MainRenderer;
		
		private Renderer.IDrawable[] DrawableElements;
		private Renderer.ILoadable[] LoadableElements;

		private Quake.VirtualMachine VM;

		private Dictionary<string, Renderer.ModelRenderer> CachedModels = new Dictionary<string, Renderer.ModelRenderer>();
		private Dictionary<string, Renderer.BspRenderer> CachedLevels = new Dictionary<string, Renderer.BspRenderer>();

		public QuakeGame() {
			this.Graphics = new GraphicsDeviceManager(this);
			this.Content = new ContentManager(Services);
			this.Graphics.PreferMultiSampling = false;
			this.Graphics.IsFullScreen = false;
			this.Graphics.SynchronizeWithVerticalRetrace = false;
			this.Graphics.PreferMultiSampling = true;
			this.Graphics.PreferredBackBufferWidth = 800;
			this.Graphics.PreferredBackBufferHeight = 600;
		}


		/// <summary>Initialise the game.</summary>
		protected override void Initialize() {

			this.VR = new VectorRenderer(this, this.Content, "LineDrawing");
			this.Components.Add(VR);

			string MapFile;
			//MapFile = @"C:\Program Files\Quake\ID1\maps\paprika.bsp";
			//MapFile = @"C:\Program Files\Quake\ID1\maps\dropflor.bsp";
			MapFile = @"C:\Program Files\Quake\ID1\maps\start.bsp";
			//MapFile = @"C:\Program Files\Quake\ID1\maps\lava.bsp";

			//Quake.Files.QuakeC.SkillLevel Skill = Quake.Files.QuakeC.SkillLevel.Easy;

			//MapFile = @"C:\Program Files\Quake2\baseq2\maps\security.bsp";

			// Set up the resource manager:
			this.Resources = new id.Quake.ResourceLoader(Path.GetDirectoryName(Path.GetDirectoryName(MapFile)));
			// Load the standard packages:
			this.Resources.AddStandardPackages();

			// Load a map:
			Stream MapSource = this.Resources.OpenFile("maps/" + Path.GetFileName(MapFile));
			Quake.Files.Level Map = new Quake.Files.Level(this.Resources, MapSource);
			Map.Name = Quake.ResourceLoader.FixFilename(Path.GetFileName(MapFile));

			Renderer.BspRenderer MapRenderer = new Renderer.BspRenderer(Map);
			this.CachedLevels.Add(Map.Name, MapRenderer);

			// Set the main renderer:
			this.MainRenderer = new Renderer.RenderController();

			// Load QuakeC:
			Quake.Files.QuakeC Progs = this.Resources.LoadObject<Quake.Files.QuakeC>("progs.dat");
			this.VM = new Quake.VirtualMachine(this.Resources, Progs);
			Quake.Files.Configuration Config = this.Resources.LoadObject<Quake.Files.Configuration>("quake.rc");
			this.VM.RunConfigurationFile(Config);

			// Attach event handlers:
			this.VM.LightStyleChanged += new Quake.VirtualMachine.LightStyleChangedEventHandler(delegate(object sender, Quake.VirtualMachine.LightStyleChangedEventArgs e) { this.MainRenderer.LightStyles[e.Light] = e.Levels; });

			this.VM.QuakeCSource.Globals["skill"].Value = 2;
			this.VM.QuakeCSource.Globals["deathmatch"].Value = false;

			this.VM.InvokeMain();
			this.VM.LoadLevel(Map);

			// Load all models:
			foreach (Quake.Files.Model m in this.VM.GetCachedModels<Quake.Files.Model>()) {
				this.CachedModels.Add(m.Name, new XnaQuake.Renderer.ModelRenderer(m));
			}
			foreach (Quake.Files.Level m in this.VM.GetCachedModels<Quake.Files.Level>()) {
				this.CachedLevels.Add(m.Name, new XnaQuake.Renderer.BspRenderer(m));
			}
	
			// Load a skybox:
			Renderer.SkyRenderer SkyRenderer = new Renderer.SkyRenderer(Map, "unit1_");

			// Set drawable and loadable elements:
			this.LoadableElements = new Renderer.ILoadable[] { MainRenderer, MapRenderer, SkyRenderer };




			string LevelName = (string)VM.Edicts["worldspawn"].Properties["message"];
			if (!string.IsNullOrEmpty(LevelName)) {
				LevelName = " - " + LevelName;
			} else {
				LevelName = "";
			}
			this.Window.Title = Assembly.GetExecutingAssembly().GetName().Name + LevelName;


			int CdTrack = (int)(float)this.VM.Edicts["worldspawn"].Properties["sounds"];
			//CdAudio.Play(CdTrack);

			List<Renderer.IDrawable> DrawableElements = new List<Renderer.IDrawable>();
			DrawableElements.Add(SkyRenderer);
			DrawableElements.Add(MapRenderer);
			foreach (Renderer.ModelRenderer ToDraw in this.CachedModels.Values) DrawableElements.Add(ToDraw);
			foreach (Renderer.BspRenderer ToDraw in this.CachedLevels.Values) DrawableElements.Add(ToDraw);
			this.DrawableElements = DrawableElements.ToArray();

			base.Initialize();
		}

		/*

		// Spawn static entities:
		void Progs_SpawnStaticEntity(object sender, Quake.Files.QuakeC.SpawnStaticEntityEventArgs e) {
			// Grab the model name from the entity.
			string Model = e.Entity.Properties["model"].String;

			// HACK: What goes on here?
			if (Model[0] == '*') return;

			// HACK: No sprite support.
			if (Model.EndsWith(".spr")) return;

			// Get the requisite renderer:
			Renderer.ModelRenderer Renderer;
			if (!CachedModels.TryGetValue(Model, out Renderer)) throw new InvalidOperationException("Model " + Model + " not cached.");

			// Add the entity's position to the renderer:
			Renderer.EntityPositions.Add(new Renderer.EntityPosition(e.Entity.Properties["origin"].Vector, e.Entity.Properties["angles"].Vector));
		}
		*/


		/// <summary>Load graphics content.</summary>
		protected override void LoadGraphicsContent(bool loadAllContent) {
			if (loadAllContent) {
				foreach (Renderer.ILoadable ToLoad in this.LoadableElements) ToLoad.Load(this.Resources, this.Graphics, this.Content);
				foreach (Renderer.ModelRenderer ToLoad in this.CachedModels.Values) ToLoad.Load(this.Resources, this.Graphics, this.Content); 
				foreach (Renderer.BspRenderer ToLoad in this.CachedLevels.Values) ToLoad.Load(this.Resources, this.Graphics, this.Content);

				this.MainRenderer.VectorRenderer = this.VR;

				// Position camera:
				Quake.Edict Player1Start = this.VM.Edicts.GetEdictByClassName("info_player_start")[0];
				MainRenderer.CameraPosition = Utility.ConvertPosition((Quake.Vector3)Player1Start.Properties["origin"]);
				MainRenderer.CameraAngle = Utility.ConvertAngle((Quake.Vector3)Player1Start.Properties["angles"]);
			}
			GC.Collect();
		}


		/// <summary>Unload graphics content.</summary>
		protected override void UnloadGraphicsContent(bool unloadAllContent) {
			if (unloadAllContent) {
				Content.Unload();
				foreach (Renderer.ILoadable ToUnload in this.LoadableElements) ToUnload.Unload(this.Graphics, this.Content);
				foreach (Renderer.ModelRenderer ToUnload in this.CachedModels.Values) ToUnload.Unload(this.Graphics, this.Content);
			}

		}


		KeyboardState LastFrameKeyboard;

		/// <summary>
		/// Allows the game to run logic such as updating the world,
		/// checking for collisions, gathering input and playing audio.
		/// </summary>
		/// <param name="gameTime">Provides a snapshot of timing values.</param>
		protected override void Update(GameTime gameTime) {

			// QuakeC logic

			this.VM.Update((float)gameTime.TotalGameTime.TotalSeconds);

			float CurrentTime = (float)gameTime.TotalGameTime.TotalSeconds;
			this.VM.QuakeCSource.Globals["time"].Value = CurrentTime;
			foreach (Quake.Edict Ed in this.VM.GetDynamicEdicts()) {
				float NextThink = (float)Ed.Properties["nextthink"];
				if (NextThink > 0f && NextThink < CurrentTime) {
					Ed.Properties["nextthink"] = 0;
					this.VM.Invoke(Ed, (Quake.Files.QuakeC.Function)Ed.Properties["think"]);
				}
			}

			// Allows the game to exit
			if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
				this.Exit();

			KeyboardState Key = Keyboard.GetState();

			if (Key.IsKeyDown(Keys.I)) {
				Quake.Edict[] Intermission = this.VM.Edicts.GetEdictByClassName("info_intermission");
				if (Intermission.Length > 0) {
					this.MainRenderer.CameraPosition = Utility.ConvertPosition((Quake.Vector3)Intermission[0].Properties["origin"]);
					this.MainRenderer.CameraAngle = Utility.ConvertAngle((Quake.Vector3)Intermission[0].Properties["mangle"]);
					CdAudio.Play(3);
				}
			}

			if (Key.IsKeyDown(Keys.Escape)) this.Exit();

			if (Key.IsKeyDown(Keys.LeftAlt) && Key.IsKeyDown(Keys.Enter) && LastFrameKeyboard.IsKeyUp(Keys.Enter)) {
				this.Graphics.ToggleFullScreen();
			}

			#region Crappy Camera Crap


			float MovementSpeed = 0f;
			if (Key.IsKeyDown(Keys.Up) && Key.IsKeyUp(Keys.Down)) {
				MovementSpeed = +1f;
			} else if (Key.IsKeyDown(Keys.Down) && Key.IsKeyUp(Keys.Up)) {
				MovementSpeed = -1f;
			}
			Vector3 NewPosition = MainRenderer.CameraPosition;



			Vector3 MovementDirection = (float)gameTime.ElapsedGameTime.TotalMilliseconds * MovementSpeed * Vector3.Transform(Vector3.Forward, Matrix.CreateFromYawPitchRoll(MainRenderer.CameraAngle.X, MainRenderer.CameraAngle.Y, MainRenderer.CameraPosition.Z)) * 0.2f;
			NewPosition += MovementDirection;

			if (Key.IsKeyDown(Keys.A)) NewPosition.Y += .2f * (float)gameTime.ElapsedGameTime.TotalMilliseconds;
			if (Key.IsKeyDown(Keys.Z)) NewPosition.Y -= .2f * (float)gameTime.ElapsedGameTime.TotalMilliseconds;

			MainRenderer.CameraPosition = NewPosition;

			Vector3 NewAngle = MainRenderer.CameraAngle;
			Vector3 AngleSpeed = Vector3.Zero;
			if (Key.IsKeyDown(Keys.Left) && Key.IsKeyUp(Keys.Right)) {
				AngleSpeed.X = +1f;
			} else if (Key.IsKeyDown(Keys.Right) && Key.IsKeyUp(Keys.Left)) {
				AngleSpeed.X = -1f;
			}

			if (Key.IsKeyDown(Keys.PageUp) && Key.IsKeyUp(Keys.PageDown)) {
				AngleSpeed.Y = +1f;
			} else if (Key.IsKeyDown(Keys.PageDown) && Key.IsKeyUp(Keys.PageUp)) {
				AngleSpeed.Y = -1f;
			}


			NewAngle += AngleSpeed * 2f * (float)gameTime.ElapsedGameTime.TotalSeconds;
			MainRenderer.CameraAngle = NewAngle;

			if (Key.IsKeyDown(Keys.R)) {
				File.WriteAllText("CAM.POS", this.MainRenderer.CameraPosition.X + ":" +
				this.MainRenderer.CameraPosition.Y + ":" +
				this.MainRenderer.CameraPosition.Z + ":" +
				this.MainRenderer.CameraAngle.X + ":" +
				this.MainRenderer.CameraAngle.Y + ":" +
				this.MainRenderer.CameraAngle.Z);
			}
			if (Key.IsKeyDown(Keys.T)) {
				RecallLastPosition();
			}

			if (Key.IsKeyDown(Keys.LeftControl)) {
				//BounceTestPosition = Utility.ConvertPosition(this.MainRenderer.CameraPosition);
				Quake.Vector3 BounceOrigin = Utility.ConvertPosition(this.MainRenderer.CameraPosition);
				this.BouncingBoundingBox = new id.Quake.BoundingBox(BounceOrigin - new Quake.Vector3(8f), BounceOrigin + new Quake.Vector3(8f));
				Quake.Vector3 CameraAngle = Utility.ConvertAngle(this.MainRenderer.CameraAngle);
				Quake.Vector3 Forwards, Right, Up;
				Quake.Vector3.GetAngleVectors(CameraAngle, out Forwards, out Right, out Up);
				BounceTestVelocity = Forwards * 300f;
				BouncePositions = new List<Vector3>();
				BouncePositions.Add(Utility.ConvertPosition(BouncingBoundingBox.MidPoint));
				BounceColours = new List<Color>();
				BounceColours.Add(Color.Red);
				
			}

			#endregion

			UpdateBounce((float)gameTime.ElapsedGameTime.TotalSeconds);

			LastFrameKeyboard = Key;
			base.Update(gameTime);
		}

		private void RecallLastPosition() {
			string[] Parms = File.ReadAllText("CAM.POS").Split(':');
			this.MainRenderer.CameraPosition = new Vector3(float.Parse(Parms[0]), float.Parse(Parms[1]), float.Parse(Parms[2]));
			this.MainRenderer.CameraAngle = new Vector3(float.Parse(Parms[3]), float.Parse(Parms[4]), float.Parse(Parms[5]));
		}


		private Quake.BoundingBox BouncingBoundingBox;
		//private Quake.Vector3 BounceTestPosition;
		private Quake.Vector3 BounceTestVelocity;
		private Quake.Vector3 BounceLastTracePosition;
		private Random RandomColourSource = new Random();

		private void UpdateBounce(float time) {
			//this.Window.Title = this.VM.Level.Mo.GetLeafFromPoint(Utility.ConvertPosition(this.MainRenderer.CameraPosition)).ToString();

			// Move the item and clip it against the world geometry.
			BounceTestVelocity.Z -= time * 200;
			//Quake.Vector3 DestinationPosition = BounceTestPosition + BounceTestVelocity * time;
			Quake.CollisionResult<Quake.BoundingBox> C = this.VM.GetCollision(BouncingBoundingBox, BounceTestVelocity * time);
			BouncingBoundingBox = C.Position;

			if (C.CollisionOccured || ((Quake.Vector3)(BouncingBoundingBox.MidPoint - BounceLastTracePosition)).Magnitude > 5) {
				BouncePositions.Add(Utility.ConvertPosition(BouncingBoundingBox.MidPoint));
				BounceColours.Add((C.CollisionOccured || BounceColours.Count < 1) ? new Color(new Vector3((float)RandomColourSource.NextDouble(), (float)RandomColourSource.NextDouble(), (float)RandomColourSource.NextDouble())) : BounceColours[BounceColours.Count - 1]);
				BounceLastTracePosition = BouncingBoundingBox.MidPoint;
			}

			if (C.CollisionOccured) {
				BounceTestVelocity = 0.7f * Quake.Vector3.Reflect(BounceTestVelocity, C.Plane.Value.Normal);
				Quake.Vector3 B = BounceTestVelocity;
				B.Normalise();
				BouncingBoundingBox += B * 0.02f;
			}


			/*Quake.Vector3 IntersectionPosition = C.Position;


			BounceTestPosition = IntersectionPosition;

			if (C.CollisionOccured || ((Quake.Vector3)(BounceTestPosition - BounceLastTracePosition)).Magnitude > 5) {
				BouncePositions.Add(Utility.ConvertPosition(C.Position));
				BounceColours.Add((C.CollisionOccured || BounceColours.Count < 1) ? new Color(new Vector3((float)RandomColourSource.NextDouble(), (float)RandomColourSource.NextDouble(), (float)RandomColourSource.NextDouble())) : BounceColours[BounceColours.Count - 1]);
				BounceLastTracePosition = C.Position;
			}

			

			if (C.CollisionOccured) {
				// Bounce!
				BounceTestVelocity = 0.7f * Quake.Vector3.Reflect(BounceTestVelocity, C.Plane.Value.Normal);
				Quake.Vector3 B = BounceTestVelocity;
				B.Normalise();
				BounceTestPosition = IntersectionPosition + B * .1f;
			}*/
		}

		private List<Vector3> BouncePositions = new List<Vector3>();
		private List<Color> BounceColours = new List<Color>();

		/// <summary>This is called when the game should draw itself.</summary>
		/// <param name="gameTime">Provides a snapshot of timing values.</param>
		protected override void Draw(GameTime gameTime) {

			int NumBalls = 0;
			Quake.Edict[] FireBalls = new id.Quake.Edict[4];
			foreach (Quake.Edict Ed in this.VM.GetDynamicEdicts()) {
				if ((string)Ed.Properties["model"] == "progs/lavaball.mdl") {
					FireBalls[NumBalls++] = Ed;
					if (NumBalls == 4) break;
				}
			}

			for (int i = 0; i < 4; ++i) {
				this.MainRenderer.WorldEffect.Parameters["Light" + i + "Brightness"].SetValue(0f);
			}

			for (int i = 0; i < NumBalls; ++i) {
				this.MainRenderer.WorldEffect.Parameters["Light" + i + "Brightness"].SetValue(8f);
				this.MainRenderer.WorldEffect.Parameters["Light" + i + "Position"].SetValue(Utility.ConvertPosition((Quake.Vector3)FireBalls[i].Properties["origin"]));
			}

			
			

			/*Vector3 FishEyeAngle = Vector3.Zero;

			Viewport V = this.Graphics.GraphicsDevice.Viewport;
			V.Width = 256;
			V.Height = 256;
			V.X = 0;
			V.Y = 256;
			this.Graphics.GraphicsDevice.Viewport = V;
			this.MainRenderer.UpdateProjection();



			this.MainRenderer.FishEyeCamera = Matrix.CreateRotationY(MathHelper.ToRadians(-90f));
			this.MainRenderer.Draw(gameTime, this.DrawableElements);

			V.X += 256; this.Graphics.GraphicsDevice.Viewport = V;
			this.MainRenderer.FishEyeCamera = Matrix.CreateRotationY(MathHelper.ToRadians(0f));
			this.MainRenderer.Draw(gameTime, this.DrawableElements);

			V.X += 256; this.Graphics.GraphicsDevice.Viewport = V;
			this.MainRenderer.FishEyeCamera = Matrix.CreateRotationY(MathHelper.ToRadians(+90f));
			this.MainRenderer.Draw(gameTime, this.DrawableElements);

			V.X += 256; this.Graphics.GraphicsDevice.Viewport = V;
			this.MainRenderer.FishEyeCamera = Matrix.CreateRotationY(MathHelper.ToRadians(+180f));
			this.MainRenderer.Draw(gameTime, this.DrawableElements);


			V.X = 256; V.Y = 0; this.Graphics.GraphicsDevice.Viewport = V;
			this.MainRenderer.FishEyeCamera = Matrix.CreateRotationX(MathHelper.ToRadians(-90f));
			this.MainRenderer.Draw(gameTime, this.DrawableElements);

			V.X = 256; V.Y = 512; this.Graphics.GraphicsDevice.Viewport = V;
			this.MainRenderer.FishEyeCamera = Matrix.CreateRotationX(MathHelper.ToRadians(+90f));
			this.MainRenderer.Draw(gameTime, this.DrawableElements);
			*/



			
			foreach (Renderer.ModelRenderer R in this.CachedModels.Values) { R.EntityPositions.Clear(); }
			foreach (Renderer.BspRenderer R in this.CachedLevels.Values) { R.EntityPositions.Clear(); }

			List<KeyValuePair<Color, BoundingBox>> BoundingBoxes = new List<KeyValuePair<Color, BoundingBox>>();

			foreach (Quake.Edict Ed in this.VM.GetDynamicEdicts()) {

				string M = (string)Ed.Properties["model"];
				XnaQuake.Renderer.EntityPosition EP = new XnaQuake.Renderer.EntityPosition((Quake.Vector3)Ed.Properties["origin"], (Quake.Vector3)Ed.Properties["angles"], new Quake.BoundingBox((Quake.Vector3)Ed.Properties["mins"], (Quake.Vector3)Ed.Properties["maxs"]));

				Color C = Color.Red;

				switch (Path.GetExtension(M)) {
					case ".mdl":
						C = Color.Lime;
						this.CachedModels[M].EntityPositions.Add(EP);
						break;
					case ".bsp":
						this.CachedLevels[M].EntityPositions.Add(EP);
						break;
				}

				BoundingBoxes.Add(new KeyValuePair<Color, BoundingBox>(C, Utility.ConvertPosition(Utility.ConvertPosition(EP.Bounds) + Utility.ConvertPosition(EP.Position))));
				
			}

			/*id.Quake.Vector3 CameraPosition = Utility.ConvertPosition(this.MainRenderer.CameraPosition);
			id.Quake.Vector3 CameraAngle = Utility.ConvertAngle(this.MainRenderer.CameraAngle);
			id.Quake.Vector3 Forwards, Right, Up;
			
			id.Quake.Vector3.GetAngleVectors(CameraAngle, out Forwards, out Right, out Up);*/


			//this.CachedModels["progs/lavaball.mdl"].EntityPositions.Add(new XnaQuake.Renderer.EntityPosition(this.BounceTestPosition, Quake.Vector3.Zero));


			this.MainRenderer.Draw(gameTime, this.DrawableElements);
			
			this.VR.SetWorldMatrix(this.MainRenderer.BaseWorldViewProj);
			

			foreach (KeyValuePair<Color, BoundingBox> box in BoundingBoxes) {
				this.VR.SetColor(box.Key);
				//this.VR.DrawBoundingBox(box.Value);				
			}


			//this.Window.Title = this.VM.Level.GetLeafFromPoint(Utility.ConvertPosition(this.MainRenderer.CameraPosition)).ToString();


			//this.VR.SetOrthogonalProjectionMatrix();
			this.VR.SetWorldMatrix(this.MainRenderer.BaseWorldViewProj);
			
			for (int i = 0; i < BouncePositions.Count - 1; ++i) {
				this.VR.SetColor(BounceColours[i]);
				//this.VR.DrawLine(BouncePositions[i + 0], BouncePositions[i + 1]);
			}

			this.VR.SetColor(Color.LimeGreen);
			this.VR.DrawBoundingBox(Utility.ConvertPosition(this.BouncingBoundingBox));
			
			base.Draw(gameTime);
		}

	}
}
