using System;
using Microsoft.Xna.Framework.Graphics;
using System.Collections.Generic;

using Quake = id.Quake;
using Xna = Microsoft.Xna.Framework;
using System.Runtime.InteropServices;
using System.IO;

namespace XnaQuake.Renderer {
	class BspRenderer : IDrawable, ILoadable {

		#region Types

		[StructLayout(LayoutKind.Sequential)]
		public struct VertexPositionTextureDiffuseLightMap {

			public Xna.Vector3 Position;
			public Xna.Vector2 DiffuseTextureCoordinate;
			public Xna.Vector2 LightMapTextureCoordinate;
			
			public VertexPositionTextureDiffuseLightMap(Xna.Vector3 position, Xna.Vector2 diffuse, Xna.Vector2 lightMap) {
				this.Position = position;
				this.DiffuseTextureCoordinate = diffuse;
				this.LightMapTextureCoordinate = lightMap;
			}

			public readonly static VertexElement[] VertexElements = new VertexElement[]{
				new VertexElement(0, 0, VertexElementFormat.Vector3, VertexElementMethod.Default, VertexElementUsage.Position, 0),
				new VertexElement(0, 12, VertexElementFormat.Vector2, VertexElementMethod.Default, VertexElementUsage.TextureCoordinate, 0),
				new VertexElement(0, 20, VertexElementFormat.Vector2, VertexElementMethod.Default, VertexElementUsage.TextureCoordinate, 1),
			};
		}

		public class Model : IDrawable {

			#region Types

			public class Leaf : IDrawable {


				private readonly Quake.Files.Level.Leaf QuakeLeaf;

				private int[] visibleLeaves = null;
				public int[] VisibleLeaves {
					get {
						if (this.visibleLeaves == null) this.visibleLeaves = this.QuakeLeaf.GetVisibilityList(this.Model.SourceModel);
						return this.visibleLeaves;
					}
				}

				public bool HasVisibilityList {
					get { return this.QuakeLeaf.HasVisibilityList; }
				}

				private readonly Model Model;

				public Leaf(BspRenderer.Model model, Quake.Files.Level.Leaf leaf) {
					this.QuakeLeaf = leaf;
					this.Model = model;
				}

				public Quake.Files.Level.Leaf.Quake2Types Type {
					get { return this.QuakeLeaf.Type; }
				}


				public override string ToString() {
					return this.QuakeLeaf.ToString();
				}


				public void Draw(RenderController controller) {
					uint FaceIndex = this.QuakeLeaf.FaceListIndex;
					for (int i = 0; i < this.QuakeLeaf.FaceCount; ++i) {
						int ActualFaceIndex = this.Model.Renderer.BspLevel.FaceList[FaceIndex++];
						this.Model.Faces[ActualFaceIndex].Draw(controller);
					}
				}
			}


			/// <summary>Represents a single face in the 3D world.</summary>
			public class Face : IDrawable {

				public int VertexIndex;
				public int VertexCount;
				public int TriangleIndex;
				public int TriangleCount;
				
				public int LightStyle;

				public bool IsLiquid;

				public AnimatedTexture Diffuse;
				public Texture2D LightMap;

				public float BaseLightLevel;

				private Model Model;


				public Face(Model model) {
					this.Model = model;
				}

				public int FrameNumber;

				public void Draw(RenderController controller) {

					//if (this.LightStyle == 0) return;
					
					if (this.FrameNumber == controller.FrameNumber) return;
					this.FrameNumber = controller.FrameNumber;
					
					if (this.Diffuse == null) return;

					controller.WorldEffect.Parameters["DiffuseTexture"].SetValue(this.Diffuse.CurrentFrame);
					controller.WorldEffect.Parameters["LightMapTexture"].SetValue(this.LightMap);
					controller.WorldEffect.Parameters["Rippling"].SetValue(this.IsLiquid);

					controller.WorldEffect.Parameters["BaseLightLevel"].SetValue(this.BaseLightLevel);

					if (this.LightStyle == 255) {
						controller.WorldEffect.Parameters["LightLevel"].SetValue(0.5f);
					} else {
						controller.WorldEffect.Parameters["LightLevel"].SetValue(controller.CurrentLightStyleLevels[this.LightStyle]);
					}

					controller.WorldEffect.CommitChanges();

					controller.Graphics.GraphicsDevice.DrawUserIndexedPrimitives<VertexPositionTextureDiffuseLightMap>(
						PrimitiveType.TriangleList,
						Model.DiffuseVertices,
						this.VertexIndex, this.VertexCount, Model.TriangleList, this.TriangleIndex, this.TriangleCount
					);
				}
			}

			public class Node  {				

				private readonly Model Model;

				public Xna.Plane Partition;

				private Node frontNode, backNode;
				private Leaf frontLeaf, backLeaf;

				private readonly Quake.Files.Level.Node SourceNode;

				public Node FrontNode {
					get {
						if (this.frontNode == null && this.SourceNode.FrontIsNode) this.frontNode = new Node(this.Model, this.Model.Renderer.BspLevel.Nodes[this.SourceNode.FrontIndex]);
						return this.frontNode;
					}
				}
				public Node BackNode {
					get {
						if (this.backNode == null && this.SourceNode.BackIsNode) this.backNode = new Node(this.Model, this.Model.Renderer.BspLevel.Nodes[this.SourceNode.BackIndex]);
						return this.backNode;
					}
				}
				public Leaf FrontLeaf {
					get {
						if (this.frontLeaf == null && !this.SourceNode.FrontIsNode) this.frontLeaf = this.Model.Leaves[this.SourceNode.FrontIndex];
						return this.frontLeaf;
					}
				}
				public Leaf BackLeaf {
					get {
						if (this.backLeaf == null && !this.SourceNode.BackIsNode) this.backLeaf = this.Model.Leaves[this.SourceNode.BackIndex];
						return this.backLeaf;
					}
				}

				public IDrawable FrontDrawable { get { return (IDrawable)this.FrontLeaf ?? (IDrawable)this.FrontNode; } }
				public IDrawable BackDrawable { get { return (IDrawable)this.BackLeaf ?? (IDrawable)this.BackNode; } }

				public Node(BspRenderer.Model model, Quake.Files.Level.Node node) {
					this.Model = model;
					this.SourceNode = node;
					this.Partition = Utility.ConvertPlane(this.Model.Renderer.BspLevel.Planes[node.Plane]);
				}

			}

			#endregion


			#region Private Members

			/// <summary>Owning renderer.</summary>
			private readonly BspRenderer Renderer;
			/// <summary>Triangle list for the entire model.</summary>
			private short[] TriangleList;
			/// <summary>All vertices with texture coordinates set up for diffuse textures.</summary>
			private VertexPositionTextureDiffuseLightMap[] DiffuseVertices;
			/// <summary>All faces that make up this model.</summary>
			public readonly Face[] Faces;
			/// <summary>Root BSP node.</summary>
			private readonly Node RootNode;
			/// <summary>The original model from Quake's BSP.</summary>
			private readonly Quake.Files.Level.Model SourceModel;

			private readonly Leaf[] Leaves;

			#endregion


			public Model(BspRenderer renderer, Quake.Files.Level.Model model) {
				this.Renderer = renderer;
				this.SourceModel = model;

				#region Load all Faces

				this.Faces = new Face[model.FaceCount];

				// Count how many vertices/triangles are required:
				int VertexCount = 0;
				int TriangleCount = 0;
				foreach (id.Quake.Files.Level.Face Face in this.Renderer.BspLevel.Faces) {
					VertexCount += Face.EdgeCount;
					TriangleCount += Face.EdgeCount - 2;
				}

				this.TriangleList = new short[TriangleCount * 3];
				this.DiffuseVertices = new VertexPositionTextureDiffuseLightMap[VertexCount];

				VertexCount = 0; TriangleCount = 0;

				// Cycle through each face in the level:
				for (int FaceIndex = 0; FaceIndex < this.Faces.Length; ++FaceIndex) {
					id.Quake.Files.Level.Face QuakeFace = this.Renderer.BspLevel.Faces[FaceIndex + model.FaceIndex];
					Face XnaFace = new Face(this); this.Faces[FaceIndex] = XnaFace;

					// Calculate the face geometry (point/triangle lists).
					#region Geometry

					// Create a Face class describing the geometry (primitives).
					XnaFace.VertexIndex = VertexCount;
					XnaFace.VertexCount = QuakeFace.EdgeCount;
					XnaFace.TriangleIndex = TriangleCount;
					XnaFace.TriangleCount = QuakeFace.EdgeCount - 2;

					// Assign the texture:
					XnaFace.Diffuse = this.Renderer.DiffuseTextures[QuakeFace.Texture];
					XnaFace.IsLiquid = this.Renderer.BspLevel.TextureInformation[QuakeFace.Texture].IsLiquid;

					XnaFace.LightStyle = QuakeFace.LightStyle;
					XnaFace.BaseLightLevel = QuakeFace.Darkness / 255f;

					// Populate the triangle list and point list with meaningful data:

					// Get the indices of the vertices that surround this face (via its edges):
					ushort[] ConvexFaceIndices = QuakeFace.GetVertexIndices();

					// TextureInfo is needed to calculate the texture coordinates.
					id.Quake.Files.Level.TextureInfo TextureInfo = this.Renderer.BspLevel.TextureInformation[QuakeFace.Texture];

					// Load lightmap texture:
					id.Quake.Files.Level.Face.LightMap LightMap = null;
					LightMap = QuakeFace.GetLightMap();
					XnaFace.LightMap = Utility.TextureFromPicture(this.Renderer.Graphics.GraphicsDevice, LightMap.Picture.GetNearestPowerOfTwo());

					// Calculate the values of the vertices that surround this face.

					for (int i = 0; i < ConvexFaceIndices.Length; ++i) {
						id.Quake.Vector3 SourceVertex = this.Renderer.BspLevel.Vertices[ConvexFaceIndices[i]];
						this.DiffuseVertices[VertexCount++] = new VertexPositionTextureDiffuseLightMap(
							Utility.ConvertPosition(SourceVertex),
							Utility.ConvertPosition(TextureInfo.GetTextureCoordinates(SourceVertex)),
							LightMap == null ? Xna.Vector2.Zero : Utility.ConvertPosition(LightMap.GetNormalisedTextureCoordinate(SourceVertex))
						);
					}

					// Calculate the values of the triangle indices for this face.
					for (int i = 2; i < ConvexFaceIndices.Length; ++i) {
						this.TriangleList[TriangleCount++] = 0;
						this.TriangleList[TriangleCount++] = ((short)(i - 1));
						this.TriangleList[TriangleCount++] = ((short)(i - 0));
					}

					#endregion

					this.RootNode = new Node(this, this.Renderer.BspLevel.Nodes[model.BspNodeIndex]);	

				}

				#endregion

				#region Leaves

				this.Leaves = new Leaf[this.Renderer.BspLevel.Leaves.Length];
				for (int i = 0; i < this.Leaves.Length; ++i) {
					this.Leaves[i] = new Leaf(this, this.Renderer.BspLevel.Leaves[i]);
				}

				#endregion

			}

			#region IDrawable Members

			public void Draw(RenderController controller) {

				if (this.SourceModel.Origin != Quake.Vector3.Zero) throw new InvalidOperationException();

				//controller.WorldEffect.Parameters["WorldViewProj"].SetValue(Xna.Matrix.CreateTranslation(-Utility.ConvertPosition(this.SourceModel.Origin)) * controller.World * controller.View * controller.Projection);// this.SourceModel.Origin

				Xna.Vector3 CameraPosition = controller.CameraPosition;

				Leaf CameraLeaf = null;
				Node CameraNode = this.RootNode;
				while (CameraLeaf == null && CameraNode != null) {
					// Are we in front of or behind the division?
					if (Xna.Vector3.Dot(CameraNode.Partition.Normal, CameraPosition) > CameraNode.Partition.D) {
						if (CameraNode.FrontLeaf != null) {
							CameraLeaf = CameraNode.FrontLeaf;
							break;
						} else {
							CameraNode = CameraNode.FrontNode;
						}
					} else {
						if (CameraNode.BackLeaf != null) {
							CameraLeaf = CameraNode.BackLeaf;
							break;
						} else {
							CameraNode = CameraNode.BackNode;
						}
					}
				}

				if (CameraLeaf.HasVisibilityList) {
					bool InSolid = (CameraLeaf == null) || ((CameraLeaf.Type & id.Quake.Files.Level.Leaf.Quake2Types.Solid) != 0);
					if (InSolid) {
						controller.Graphics.GraphicsDevice.RenderState.FillMode = FillMode.WireFrame;
						foreach (Face ToDraw in this.Faces) ToDraw.Draw(controller);
						controller.Graphics.GraphicsDevice.RenderState.FillMode = FillMode.Solid;
					} else {
						foreach (uint i in CameraLeaf.VisibleLeaves) {
							this.Leaves[i].Draw(controller);
						}
					}
				} else {
					foreach (Face F in this.Faces) {
						F.Draw(controller);						
					}
				}
				
	
			}

			#endregion
		}

		public class AnimatedTexture  : IDisposable {
			public Texture2D CurrentFrame;
			public Texture2D[] Frames;
			private int CurrentFrameIndex = -1;
			public void Advance() {
				this.CurrentFrameIndex = (this.CurrentFrameIndex + 1) % this.Frames.Length;
				this.CurrentFrame = this.Frames[this.CurrentFrameIndex];
			}
			public void Dispose() {
				foreach (Texture2D T in this.Frames) {
					if (T != null && T.IsDisposed) T.Dispose();
				}
			}
		}

		#endregion

		#region Private Members

		/// <summary>The BSP level that this renderer has to draw.</summary>
		private readonly id.Quake.Files.Level BspLevel;

		private AnimatedTexture[] DiffuseTextures;

		private AnimatedTexture[] TexturesToAnimate;

		#endregion

		#region Public Members

		public List<EntityPosition> EntityPositions = new List<EntityPosition>();

		#endregion

		#region Public Properties

		private Model[] models;
		/// <summary>Gets the models that make up the BSP level.</summary>
		public Model[] Models {
			get { return this.models; }
		}

		#endregion

		#region Constructor

		public BspRenderer(id.Quake.Files.Level level) {
			this.BspLevel = level;
		}

		#endregion

		#region ILoadable Members

		private Xna.GraphicsDeviceManager Graphics;

		public void Load(id.Quake.ResourceLoader resources, Microsoft.Xna.Framework.GraphicsDeviceManager graphics, Xna.Content.ContentManager content) {

			this.Graphics = graphics;

			// Load all diffuse textures:

			this.DiffuseTextures = new AnimatedTexture[this.BspLevel.TextureInformation.Length];
			string[] DiffuseTextureNames = new string[this.DiffuseTextures.Length];

			Dictionary<string, int> AnimatedTextureCache = new Dictionary<string, int>(this.BspLevel.TextureInformation.Length);

			for (int i = 0; i < this.BspLevel.TextureInformation.Length; ++i) {
				Quake.Files.Level.TextureInfo Texture = this.BspLevel.TextureInformation[i];
				if (Texture.IsVisible && !Texture.IsSky) {
					int j;
					if (AnimatedTextureCache.TryGetValue(Texture.Name, out j)) {
						DiffuseTextures[i] = DiffuseTextures[j];
					} else {
						DiffuseTextures[i] = new AnimatedTexture();
						DiffuseTextures[i].Frames = new Texture2D[] { Utility.TextureFromMipMappedPicture(graphics.GraphicsDevice, Texture) };
						AnimatedTextureCache.Add(Texture.Name, i);
					}
				}
			}

			// Search for all animated textures:
			Dictionary<string, string> NextFrameNames = new Dictionary<string, string>(this.BspLevel.TextureInformation.Length);
			for (int i = 0; i < this.BspLevel.TextureInformation.Length; ++i) {
				Quake.Files.Level.TextureInfo Texture = this.BspLevel.TextureInformation[i];
				if (!NextFrameNames.ContainsKey(Texture.Name)) {
					string NextName = Texture.NextName;
					if (!string.IsNullOrEmpty(NextName)) NextFrameNames.Add(Texture.Name, NextName);
				}
			}

			// Clump animated textures into sequences.
			List<AnimatedTexture> TexturesToAnimate = new List<AnimatedTexture>();
			while (NextFrameNames.Count > 0) {
				List<string> AnimatedSequence = new List<string>(NextFrameNames.Count);
				IEnumerator<KeyValuePair<string, string>> GetFirstName = NextFrameNames.GetEnumerator();
				GetFirstName.MoveNext();
				string CurrentFrame = GetFirstName.Current.Key;
				while (NextFrameNames.ContainsKey(CurrentFrame)) {
					AnimatedSequence.Add(CurrentFrame);
					string NextFrame = NextFrameNames[CurrentFrame];
					NextFrameNames.Remove(CurrentFrame);
					CurrentFrame = NextFrame;
				}

				// AnimatedSequence is now a list of texture names in animated sequence.
				for (int i = 0; i < AnimatedSequence.Count; ++i) {
					int AnimatedTextureIndex = AnimatedTextureCache[AnimatedSequence[i]];
					TexturesToAnimate.Add(this.DiffuseTextures[AnimatedTextureIndex]);
					Array.Resize(ref this.DiffuseTextures[AnimatedTextureIndex].Frames, AnimatedSequence.Count);
					for (int j = 1; j < AnimatedSequence.Count; ++j) {
						this.DiffuseTextures[AnimatedTextureIndex].Frames[j] =
							this.DiffuseTextures[AnimatedTextureCache[AnimatedSequence[(i + j) % AnimatedSequence.Count]]].Frames[0];
					}
				}
			}
			this.TexturesToAnimate = TexturesToAnimate.ToArray();

			// Prime all textures:
			foreach (AnimatedTexture AT in this.DiffuseTextures) {
				if (AT != null) AT.Advance();
			}
			
			#region Models

			this.models = new Model[this.BspLevel.Models.Length];
			for (int i = 0; i < this.models.Length; ++i) {
				this.models[i] = new Model(this, this.BspLevel.Models[i]);
			}

			#endregion
		}



		#endregion

		#region IDrawable Members

		double TextureAnimationTime = 0;
		public void Draw(RenderController controller) {

			TextureAnimationTime += controller.Time.ElapsedGameTime.TotalSeconds * 5d;
			while (TextureAnimationTime > 0) {
				foreach (AnimatedTexture AT in TexturesToAnimate) AT.Advance();
				TextureAnimationTime -= 1d;
			}

			controller.WorldEffect.CurrentTechnique = controller.WorldEffect.Techniques["World"];
			
			controller.WorldEffect.Begin();

			controller.VertexType = VertexPositionTextureDiffuseLightMap.VertexElements;

			foreach (EffectPass Pass in controller.WorldEffect.CurrentTechnique.Passes) {
				Pass.Begin();
				foreach (EntityPosition Pos in this.EntityPositions) {
					controller.WorldEffect.Parameters["WorldViewProj"].SetValue(Xna.Matrix.CreateFromYawPitchRoll(Pos.Angle.X, Pos.Angle.Y, Pos.Angle.Z) * Xna.Matrix.CreateTranslation(Pos.Position) * controller.BaseWorldViewProj);
					controller.WorldEffect.CommitChanges();					
				}
				foreach (Model ToDraw in this.Models) {
					ToDraw.Draw(controller);
				}
				Pass.End();
			}

			controller.WorldEffect.End();

			
		}

		#endregion

		#region ILoadable Members


		public void Unload(Microsoft.Xna.Framework.GraphicsDeviceManager graphics, Microsoft.Xna.Framework.Content.ContentManager content) {
			foreach (Model M in this.Models) {
				foreach (Model.Face F in M.Faces) {
					if (F.Diffuse != null) F.Diffuse.Dispose();
					if (F.LightMap != null) F.LightMap.Dispose();
				}				
			}
		}

		#endregion
	}
}
